Package: iscream
Title: Make fast and memory efficient BED file queries, summaries and
        matrices
Version: 1.0.0
Authors@R: c(
    person("James", "Eapen", email = "james.eapen@vai.org", role = c("aut", "cre"), comment = c(ORCID = "0000-0001-6016-3598")),
    person("Jacob", "Morrison", email = "jacob.morrison@vai.org", role = c("aut"), comment = c(ORCID = "0000-0001-8592-4744")),
    person("Nathan", "Spix", role = c("ctb")),
    person("Hui", "Shen", email = "hui.shen@vai.org", role = c("aut", "ths", "fnd"), comment = c(ORCID = "0000-0001-9767-4084")))
Description: BED files store ranged genomic data that can be queried even when
             the files are compressed. iscream can query data from BED files and return them
             in muliple formats: parsed records or their summary statistics as data frames
             or GenomicRanges objects, and matrices as matrix, GenomicRanges, or
             SummarizedExperiment objects. iscream also provides specialized
             support for importing methylation data.
URL: https://huishenlab.github.io/iscream/,
        https://github.com/huishenlab/iscream/
BugReports: https://github.com/huishenlab/iscream/issues/
License: MIT + file LICENSE
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
Depends: R (>= 4.5)
LinkingTo: Rcpp, RcppArmadillo, RcppProgress, RcppSpdlog, Rhtslib,
        stringfish
Imports: Rcpp, Matrix, data.table, methods, pbapply, parallelly,
        stringfish,
Suggests: BiocFileCache, BiocStyle, bsseq, ggplot2, ggridges, knitr,
        microbenchmark, rmarkdown, GenomicRanges, IRanges, Rsamtools,
        SummarizedExperiment, S4Vectors, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: yes
SystemRequirements: htslib: htslib-devel (rpm) or libhts-dev (deb) &
        tabix: htslib-tools (rpm) or tabix (deb) & GNU make
biocViews: DataImport, Software, Sequencing, SingleCell, DNAMethylation
git_url: https://git.bioconductor.org/packages/iscream
git_branch: RELEASE_3_22
git_last_commit: faf115f
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
Packaged: 2025-10-30 04:39:58 UTC; biocbuild
Author: James Eapen [aut, cre] (ORCID: <https://orcid.org/0000-0001-6016-3598>),
  Jacob Morrison [aut] (ORCID: <https://orcid.org/0000-0001-8592-4744>),
  Nathan Spix [ctb],
  Hui Shen [aut, ths, fnd] (ORCID:
    <https://orcid.org/0000-0001-9767-4084>)
Maintainer: James Eapen <james.eapen@vai.org>

# Generated by roxygen2: do not edit by hand

export(get_df_string)
export(get_granges_string)
export(get_log_level)
export(get_threads)
export(htslib_version)
export(make_mat)
export(make_mat_bsseq)
export(make_mat_gr)
export(make_mat_se)
export(query_chroms)
export(set_log_level)
export(set_threads)
export(summarize_meth_regions)
export(summarize_regions)
export(tabix)
export(tabix_gr)
export(tabix_raw)
import(Matrix)
import(stringfish)
importFrom(Matrix,drop0)
importFrom(Rcpp,sourceCpp)
importFrom(data.table,":=")
importFrom(data.table,as.data.table)
importFrom(data.table,fread)
importFrom(data.table,fwrite)
importFrom(data.table,rbindlist)
importFrom(data.table,set)
importFrom(data.table,setDT)
importFrom(data.table,setnames)
importFrom(data.table,tstrsplit)
importFrom(methods,is)
importFrom(parallelly,availableCores)
importFrom(pbapply,pblapply)
importFrom(pbapply,pboptions)
importFrom(stats,setNames)
importFrom(stringfish,sf_grepl)
importFrom(tools,file_path_sans_ext)
useDynLib(iscream, .registration = TRUE)

## iscream <a href="https://huishenlab.github.io/iscream/"><img src="man/figures/logo.png" align="right" height="138" style="float:right; height:138px;"/></a>

<!-- badges: start -->
[![R-CMD-check-bioc](https://github.com/huishenlab/iscream/actions/workflows/check-bioc.yml/badge.svg)](https://github.com/huishenlab/iscream/actions/workflows/check-bioc.yml)
[![Codecov test coverage](https://codecov.io/gh/huishenlab/iscream/graph/badge.svg)](https://app.codecov.io/gh/huishenlab/iscream)
<!-- badges: end -->

iscream aims to efficiently read data from any BED file into formats usable by
other packages. Using [htslib](https://www.htslib.org/), iscream can query
genomic regions like [tabix](https://en.wikipedia.org/wiki/Tabix), summarize the
queried data and make matrices, with specific support for WGBS BED files aligned
by [BISCUIT](https://huishenlab.github.io/biscuit/),
[Bismark](https://www.bioinformatics.babraham.ac.uk/projects/bismark/) and
[BSBolt](https://bsbolt.readthedocs.io/en/latest/).

Analysis and visualization of Whole Genome Bisulfite Sequencing (WGBS)[^1] data
requires reading aligned sequencing data into formats that existing packages
like [BSseq](https://bioconductor.org/packages/devel/bioc/html/bsseq.html) and
[scMET](https://github.com/andreaskapou/scMET) can analyze. Getting the data
from on-disk BED files to a matrix of methylation values can be difficult
because, with nearly 30 million CpGs, WGBS data can be quite large. iscream
makes importing WGBS data for targeted exploration and analysis faster and more
memory efficient.

[^1]: The name iscream comes from "*Integrating Single-Cell Results for
Exploring and Analyzing Methylation*" as it was originally developed to read BED
files from WGBS. It was then generalized to work with any BED file.


## Dependencies

### *htslib* >= 1.17

*iscream* depends on the *htslib* header files. These may be installed with
your package manager:

- ubuntu/debian: `libhts-dev`  
- fedora/RHEL: `htslib-devel`  
- brew: `htslib`  
- nixpkgs: `htslib`
- conda: `bioconda::htslib`

or built manually: <https://www.htslib.org/download/>. *iscream* can use
*Rhtslib* as the htslib source but we recommend installing htslib with
libdeflate support for optimal performance - see
[`vignette("htslib")`](https://huishenlab.github.io/iscream/articles/htslib.html)
for more information.


The header files may also be found among your HPC modules - make sure the
`PKG_CONFIG_PATH` environment variable includes the `pkgconfig` location for
your installation of *htslib*. You can verify that the *htslib* development
libraries are installed with `pkg-config`:

```bash
# set path if necessary
export PKG_CONFIG_PATH=[path to htslib installation]
# verify that htslib can be found
pkg-config --cflags --libs htslib
```

#### *tabix*

Some *htslib* installations do not include the *tabix* executable (on Ubuntu you
need to install both *libhts-dev* and *tabix*). *iscream* will work without
*tabix*, but the `tabix()` function will be faster if the executable is
installed.

### GCC >= 9.4.0

GNU GCC must be installed for OpenMP support. This is usually installed by
default on Linux systems, but may need to be manually installed on MacOS to use
iscream with multiple threads[^2].

[^2]: Using OpenMP is also possible with Clang on MacOS
(<https://mac.r-project.org/openmp/>) but installing GCC with Homebrew may be
easier (<https://formulae.brew.sh/formula/gcc>).

## Installation

### Bioconductor

```r
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("iscream")
```

### GitHub

You can install the development version from Github by cloning the repo and
running

```bash
git clone https://github.com/huishenlab/iscream
R CMD INSTALL iscream
```

You can also use the R [`devtools`](https://devtools.r-lib.org/) package:

```r
devtools::install_github("huishenlab/iscream")
```

or [`pak`](https://pak.r-lib.org/):

```r
pak::pkg_install("huishenlab/iscream")
```

### Usage

See the [quick start guide](https://huishenlab.github.io/iscream/articles/iscream.html)
for an overview of iscream's functionality and the [function reference](https://huishenlab.github.io/iscream/reference/)
for all available functions. Bug reports may be submitted through [GitHub issues](https://github.com/huishenlab/iscream/issues).

#' @rdname tabix
#' @export
tabix_raw <- function(
  bedfiles,
  regions,
  nthreads = NULL
) {
  input_regions <- get_string_input_regions(regions)
  run_scan_tabix(bedfiles, input_regions, .get_threads(nthreads))
}

library(data.table)
tabix_installed <- Sys.which("tabix") != ""
options("tabix.method" = "htslib")

# input data
extdata <- system.file("extdata", package = "iscream")
chrom_beds <- list.files(file.path(extdata, "test_chroms"), pattern = "*.bed.gz$", full.names = T)
biscuit_tabix_beds <- list.files(extdata, pattern = "[a|b|c|d].bed.gz$", full.names = T)
bismark_tabix_beds <- list.files(extdata, pattern = "[a|b|c|d].cov.gz$", full.names = T)
mergecg_bed <- list.files(extdata, pattern = "*_mergecg.bed.gz$", full.names = T)

regions <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14")
regions.missing_in_3 <- c("chr1:13-14")
regions.dt <- as.data.table(regions)[, tstrsplit(regions, ":|-")]
colnames(regions.dt) <- c("chr", "start", "end")
chrs <- sort(paste0("chr", c(seq(1:22), "M", "X", "Y")))
gr <- GenomicRanges::GRanges(regions)

test_that("query_chroms", {
  expect_equal(
    chrs,
    query_chroms(chrom_beds)
  )
})

# testing data
tabix_df_result <- list.files(extdata, pattern = "tabix_dataframe.test", full.names = T)
tabix_df_result_bismark <- list.files(extdata, pattern = "tabix_dataframe_bismark.test", full.names = T)

# test dataframe output
test_tabix_dataframe <- function(shell = FALSE) {
  if (shell) {
    options("tabix.method" = "shell")
  }
  expect_equal(
    tabix(biscuit_tabix_beds[1], regions, aligner = "biscuit"),
    fread(tabix_df_result, colClasses = c("character", "numeric", "numeric", "numeric", "numeric"))
  )
  expect_equal(
    tabix(biscuit_tabix_beds[1], regions.dt, aligner = "biscuit"),
    fread(tabix_df_result, colClasses = c("character", "numeric", "numeric", "numeric", "numeric"))
  )
  expect_equal(
    tabix(biscuit_tabix_beds[1], gr, aligner = "biscuit"),
    fread(tabix_df_result, colClasses = c("character", "numeric", "numeric", "numeric", "numeric"))
  )
  expect_equal(
    tabix(bismark_tabix_beds[1], regions, aligner = 'bismark'),
    fread(tabix_df_result_bismark, colClasses = c("character", "numeric", "numeric", "numeric", "numeric", "numeric"))
  )
  # empty
  expect_warning(
    tabix(chrom_beds[2], regions),
    "No records found"
  )
}

test_that("tabix dataframe with shell", {
  skip_if_not(tabix_installed)
  test_tabix_dataframe(shell = TRUE)
})

test_that("tabix dataframe with htslib", {
  options("tabix.method" = "htslib")
  test_tabix_dataframe(shell = FALSE)
})

test_multi_tabix_dataframe <- function(shell = FALSE) {
  if (shell) {
    options("tabix.method" = "shell")
  }
  multi_tabix <- lapply(biscuit_tabix_beds, function(i) {
    tabix(i, regions)[,
      file := tools::file_path_sans_ext(basename(i), compression = TRUE)
    ]
  }) |>
    rbindlist()
  multi_tabix_biscuit <- lapply(biscuit_tabix_beds, function(i) {
    tabix(i, regions, aligner = "biscuit")[,
      file := tools::file_path_sans_ext(basename(i), compression = TRUE)
    ]
  }) |>
    rbindlist()
  multi_tabix_bismark <- lapply(bismark_tabix_beds, function(i) {
    tabix(i, regions, aligner = "bismark")[,
      file := tools::file_path_sans_ext(basename(i), compression = TRUE)
    ]
  }) |>
    rbindlist()
  multi_tabix_custom <- lapply(biscuit_tabix_beds, function(i) {
    tabix(i, regions, col.names = c("1", "2"))[,
      file := tools::file_path_sans_ext(basename(i), compression = TRUE)
    ]
  }) |>
    rbindlist()

  expect_equal(
    tabix(biscuit_tabix_beds, regions),
    multi_tabix
  )
  expect_equal(
    tabix(biscuit_tabix_beds, regions, aligner = "biscuit"),
    multi_tabix_biscuit
  )
  expect_equal(
    tabix(bismark_tabix_beds, regions, aligner = "bismark"),
    multi_tabix_bismark
  )
  expect_equal(
    tabix(biscuit_tabix_beds, regions, col.names = c("1", "2")),
    multi_tabix_custom
  )
  expect_equal(
    suppressWarnings(tabix(biscuit_tabix_beds, regions.missing_in_3)),
    tabix(biscuit_tabix_beds[-3], regions.missing_in_3)
  )
}

test_that("tabix multi query with shell", {
  skip_if_not(tabix_installed)
  test_multi_tabix_dataframe(shell = TRUE)
})

test_that("tabix multi query with htslib", {
  options("tabix.method" = "htslib")
  test_multi_tabix_dataframe(shell = FALSE)
})

suppressPackageStartupMessages(library(GenomicRanges))
gr.meta <- GRanges(regions)
values(gr.meta) <- data.frame(meta = c("s1", "s2", "s3"))

test_that("tabix_gr", {
  # gr input
  expect_equal(
    tabix_gr(biscuit_tabix_beds[1], gr, aligner = "biscuit"),
    fread(tabix_df_result, colClasses = c("character", "numeric", "numeric", "numeric", "numeric")) |>
      GenomicRanges::makeGRangesFromDataFrame(starts.in.df.are.0based = TRUE, keep.extra.columns = TRUE)
  )
  # string input
  expect_equal(
    tabix_gr(biscuit_tabix_beds[1], regions, aligner = "biscuit"),
    fread(tabix_df_result, colClasses = c("character", "numeric", "numeric", "numeric", "numeric")) |>
      GenomicRanges::makeGRangesFromDataFrame(starts.in.df.are.0based = TRUE, keep.extra.columns = TRUE)
  )
  # dt input
  expect_equal(
    tabix_gr(biscuit_tabix_beds[1], regions.dt, aligner = "biscuit"),
    fread(tabix_df_result, colClasses = c("character", "numeric", "numeric", "numeric", "numeric")) |>
      GenomicRanges::makeGRangesFromDataFrame(starts.in.df.are.0based = TRUE, keep.extra.columns = TRUE)
  )
  # gr input with metadata
  expect_equal(
    tabix_gr(biscuit_tabix_beds[1], gr.meta, aligner = "biscuit"),
    fread(tabix_df_result, colClasses = c("character", "numeric", "numeric", "numeric", "numeric"))[,
      meta := c(rep("s1", 3), rep("s2", 2), rep("s3", 2))
    ] |>
      GenomicRanges::makeGRangesFromDataFrame(starts.in.df.are.0based = TRUE, keep.extra.columns = TRUE)
  )
  # gr input with metadata, bismark
  expect_equal(
    tabix_gr(bismark_tabix_beds[1], gr.meta, aligner = 'bismark', zero_based = FALSE),
    fread(tabix_df_result_bismark, colClasses = c("character", "numeric", "numeric", "numeric", "numeric", "numeric"))[,
      meta := c(rep("s1", 3), rep("s2", 2), rep("s3", 2))
    ] |>
      GenomicRanges::makeGRangesFromDataFrame(starts.in.df.are.0based = FALSE, keep.extra.columns = TRUE)
  )
  expect_error(suppressWarnings(tabix_gr(biscuit_tabix_beds[1], "chrM:1-10")))
})

test_that("tabix multi GR(List)", {
  grl <- GenomicRanges::GRangesList(
    lapply(seq_len(length(biscuit_tabix_beds)), function(bed) {
      res <- tabix_gr(biscuit_tabix_beds[bed], gr)
      GenomicRanges::mcols(res)$file <- letters[bed]
      res
    })
  )
  names(grl) <- c(letters[1:4])

  res <- tabix_gr(biscuit_tabix_beds, gr)
  expect_s4_class(res, "CompressedGRangesList")
  expect_equal(names(res), letters[1:4])
  expect_equal(res, grl)
})

# test colnames
base_colnames <- c("chr", "start", "end")
custom_cols <- paste0("c", 1:2)
test_that("tabix custom colnames", {
  expect_equal(
    colnames(tabix(chrom_beds[1], regions, col.names = custom_cols)),
    c(base_colnames, custom_cols)
  )
})

custom_cols <- paste0("c", 1:6)
test_that("tabix custom colnames too many", {
  expect_warning(
    tabix(chrom_beds[1], regions, col.names = custom_cols),
    "Fewer columns in data than provided colnames"
  )
})

custom_cols <- paste0("c1")
test_that("tabix custom colnames too few", {
  expect_warning(
    tabix(chrom_beds[1], regions, col.names = custom_cols),
    paste("Did not use input data 'colnames' - only", length(custom_cols), "names provided for 2 data columns")
  )
})


biscuit_colnames <- c("beta", "coverage")
bismark_colnames <- c("methylation.percentage", "count.methylated", "count.unmethylated")
test_that("tabix mergecg colnames", {
  expect_equal(
    colnames(tabix(mergecg_bed, regions, aligner = "biscuit")),
    c(base_colnames, biscuit_colnames, "mergecg")
  )
})

# test raw output
tabix_raw_res <- list(
  `chr1:1-6` = c("chr1\t2\t4\t1.000\t2"),
  `chr1:7-10` = c("chr1\t6\t8\t0.000\t2", "chr1\t8\t10\t1.000\t1"),
  `chr1:11-14` = character(0)
)

test_that("tabix raw list", {
  expect_equal(
    tabix_raw(biscuit_tabix_beds[3], regions),
    tabix_raw_res
  )
})


multi_tabix_raw <- lapply(biscuit_tabix_beds, function(i) {
  tabix_raw(i, regions)
})
names(multi_tabix_raw) = tools::file_path_sans_ext(basename(biscuit_tabix_beds), compression = TRUE)

test_that("tabix multi query raw", {
  expect_equal(
    tabix_raw(biscuit_tabix_beds, regions),
    multi_tabix_raw
  )
})


test_that("check tabix.method", {
  with_mocked_bindings(
    {
      tabix_msg <- "tabix' executable not found in $PATH. tabix() will use htslib to make queries instead which can be slower. See ?tabix for details."
      expect_true(grepl(tabix_msg, iscream:::package_loader(), fixed = TRUE))
      expect_equal(getOption("tabix.method"), "htslib")
    },
    Sys.which = function(input) "",
    .package = "base"
  )
  with_mocked_bindings(
    {
      iscream:::package_loader()
      expect_equal(getOption("tabix.method"), "shell")
    },
    Sys.which = function(input) "tabix",
    .package = "base"
  )
})

## ----echo=FALSE, error=FALSE--------------------------------------------------
if (!require("ggplot2", quietly = TRUE)) {
  stop("The 'ggplot2' package must be installed for this functionality")
}

## ----echo=FALSE, dpi=300, fig.width=6, fig.height=4, fig.align="center", fig.cap="Effect of htslib v1.18 source on iscream's 'make_mat()' runtime from one bulk WGBS BED file"----
data_dir <- system.file("extdata", package = "iscream")
df <- read.csv(file.path(data_dir, "htslib.csv"))
ggplot(df, aes(x = regions, y = time, color = source)) +
  geom_point(alpha = 0.8) +
  stat_summary(
    aes(x = regions, group = source),
    fun = "mean",
    geom = "line",
    linewidth = 0.5,
    show.legend = TRUE
  ) +
  labs(
    x = "Genomic regions",
    y = "Runtime (s)",
    color = "htslib source"
  ) +
  theme_bw()



## ----si-----------------------------------------------------------------------
sessionInfo()

## usethis namespace: start
#' @author James Eapen
#' @name iscream
#' @useDynLib iscream, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @import Matrix
#' @import stringfish
## usethis namespace: end
NULL

#' Summarize information over genomic regions from any BED file
#'
#' Run summarizing functions on BED file records across genomic regions.
#' Parallelized across files using threads from the `"iscream.threads"` option.
#' @param bedfiles A vector of BED file paths
#' @param regions A vector, data frame or GenomicRanges of genomic regions. See
#' details.
#' @param columns A vector of indices of the numeric columns to be summarized
#' @param col_names A vector of names to use for `columns` in the output
#' @param fun Function(s) to apply over the region. See details.
#' @param feature_col Column name of the input `regions` data frame containing
#' a name for each genomic region. Set only if the using a data frame as the
#' input regions format. See details.
#' @param set_region_rownames Use the region strings as the returned data
#' frame's rownames. Can be useful if you have a named regions and want both
#' the regions strings rownames and the feature names. See details.
#' @param nthreads Set number of threads to use overriding the
#' `"iscream.threads"` option. See `?set_threads` for more information.
#'
#' @details
#'
#' # Supported functions
#'
#' - Sum: `"sum"`
#' - Mean: `"mean"`
#' - Median: `"median"`
#' - Standard deviation: `"stddev"`
#' - Variance: `"variance"`
#' - Minimum: `"min"`
#' - Maximum: `"max"`
#' - Range: `"range"`
#' - No. of records in the region: `"count"`
#'
#' The summarizing computations are backed by the Armadillo library. See
#' <https://arma.sourceforge.net/docs.html#stats_fns> for futher details on the
#' supported functions
#'
#' # Using feature identifiers
#'
#' `regions` may be string vector in the form "chr:start-end", a GRanges
#' object or a data frame with "chr", "start", and "end" columns. The `feature`
#' column of the output will contain a "chr:start-end" identifier for each
#' summarized region. To use other identifiers, like a gene name for a region
#' instead of the coordinates, set the names of the vector or GRanges to those
#' identifiers. These names will be used instead of the genomic region string
#' to describe each feature in the output dataframe. If `regions` is a data
#' frame make an additional column with the identifiers and pass that column
#' name to `feature_col`. See examples.
#'
#' @returns A data.frame
#'
#' @importFrom methods is
#'
#' @export
#'
#' @examples
#' bedfiles <- system.file("extdata", package = "iscream") |>
#'   list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
#' # examine the bedfiles
#' colnames <- c("chr", "start", "end", "beta", "coverage")
#' lapply(bedfiles, function(i) knitr::kable(read.table(i, col.names = colnames)))
#'
#' # make a vector of regions
#' regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
#' summarize_regions(bedfiles, regions, columns = c(4, 5), col_names = c("beta", "cov"))
#'
#' # select functions
#' summarize_regions(
#'   bedfiles,
#'   regions,
#'   fun = c("mean", "stddev"),
#'   columns = c(4, 5),
#'   col_names = c("beta", "cov")
#' )
#'
#' # add names to the regions
#' names(regions) <- c("A", "B", "C")
#' summarize_regions(
#'   bedfiles,
#'   regions,
#'   fun = "sum",
#'   columns = 5,
#'   col_names = "coverage"
#' )
#'
#' # using `feature_col`
#' library(data.table)
#'
#' # convert string vector to a data.table
#' regions_df <- data.table::as.data.table(regions) |>
#' _[, tstrsplit(regions, ":|-", fixed = FALSE, names = c("chr", "start", "end"))] |>
#' _[, start := as.integer(start)] |>
#' _[, feature := LETTERS[.I]][]
#' regions_df
#'
#' summarize_regions(
#'   bedfiles,
#'   regions_df,
#'   fun = "sum",
#'   columns = 5,
#'   col_names = "coverage",
#'   feature_col = "feature"
#' )
summarize_regions <- function(
  bedfiles,
  regions,
  columns,
  col_names = NULL,
  fun = "all",
  feature_col = NULL,
  set_region_rownames = FALSE,
  nthreads = NULL
) {
  supported_funcs <- c("sum", "mean", "median", "stddev", "variance", "min", "max", "range", "count")

  if (length(fun) > 1) {
    if ("all" %in% fun) {
      stop("'all' can't be used with other summary funcions")
    }
    stopifnot("Selected function not supported" = all(fun %in% supported_funcs))
    fun_to_use <- fun
  } else {
    stopifnot("Selected function not supported" = fun %in% c(supported_funcs, "all"))
    fun_to_use <- supported_funcs
    if (fun != "all") {
      fun_to_use <- fun
    }
  }

  col_names <- col_names %||% paste0("V", seq_len(length(columns)))

  verify_files_or_stop(bedfiles, verify_tabix = TRUE)
  if (is(regions, "GRanges")) {
    regions <- get_granges_string(regions)
  } else if ("data.frame" %in% class(regions)) {
    regions <- get_df_string(regions, feature_col)
  }

  n_threads <- .get_threads(nthreads)
  validate_log_level(n_threads = n_threads)

  df <- Cpp_summarize_regions(
    bedfiles = bedfiles,
    regions = regions,
    col_indices = columns,
    col_names = col_names,
    fun_vec = fun_to_use,
    region_rownames = set_region_rownames,
    aligner = "general",
    mval = FALSE,
    nthreads = n_threads
  )
  df[df == -99] <- NA
  df

  count_colnames <- paste0(col_names, ".count")
  if (any(count_colnames %in% colnames(df))) {
    df <- df[, !(names(df) %in% count_colnames[-1])]
  }

  colnames(df)[which(colnames(df) == count_colnames[1])] <- "count"
  df
}

library("stringfish")
library("data.table")

# inputs
extdata <- system.file("extdata", package = "iscream")
biscuit_bedfiles <- list.files(extdata, pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
bismark_bedfiles <- list.files(extdata, pattern = "[a|b|c|d].cov.gz$", full.names = TRUE)
regions <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14")
regions.dt <- as.data.table(regions)[, tstrsplit(regions, ":|-")][, names := names(regions)]
colnames(regions.dt) <- c("chr", "start", "end", "names")
gr <- GenomicRanges::GRanges(regions)
supported_funcs <- c("sum", "mean", "median", "stddev", "variance", "min", "max", "range")

get_colnames <- function(funcs, col_names) {
  base_colnames <- c("feature", "file")
  values <- c(col_names)
  if ("all" %in% funcs) {
    return(c(base_colnames, as.vector(outer(values, supported_funcs, paste, sep = ".")), "count"))
  }

  if ("count" %in% funcs) {
    index <- which(funcs == "count")
    funcs <- funcs[-index]
  }
  function_used <- funcs
  col_names <- c(
    base_colnames,
    as.vector(outer(values, function_used, paste, sep = "."))
  )

  if ("count" %in% funcs) {
  col_names <- c(col_names, "count")
  }

  return(col_names)
}

run_test <- function(bedfiles, regions, funcs, columns, col_names, nthreads) {
  test_colnames <- get_colnames(funcs, col_names)
  reg_length <- if ("data.frame" %in% class(regions)) {
    nrow(regions)
  } else {
      length(regions)
  }

  list(
    df = summarize_regions(bedfiles, regions, fun = funcs, columns = columns, col_names = col_names, nthreads = nthreads),
    colnames = test_colnames,
    dims = c(length(bedfiles) * reg_length, length(test_colnames))
  )
}

one_file <- lapply(biscuit_bedfiles, function(i) {
  run_test(
    i,
    regions,
    funcs = "all",
    c(4, 5),
    c("beta", "coverage"),
    nthreads = 1
  )
})

one_fun <- lapply(supported_funcs, function(i) {
  run_test(
    biscuit_bedfiles,
    regions,
    funcs = i,
    c(4, 5),
    c("beta", "coverage"),
    nthreads = 1
  )
})

two_fun <- lapply(asplit(combn(supported_funcs, 2), 2), function(i) {
  run_test(
    biscuit_bedfiles,
    regions,
    funcs = i,
    c(4, 5),
    c("cov", "beta"),
    nthreads = 1
  )
})

one_reg <- lapply(regions, function(i) {
  run_test(
    biscuit_bedfiles,
    i,
    funcs = "all",
    c(4, 5),
    c("beta", "coverage"),
    nthreads = 1
  )
})

threads <- lapply(c(1, 1), function(i) {
  run_test(
    biscuit_bedfiles,
    regions,
    funcs = "all",
    c(4, 5),
    c("beta", "coverage"),
    nthreads = i
  )
})

regions_df <- list(run_test(
  biscuit_bedfiles,
  regions.dt,
  funcs = "all",
  c(4, 5),
  c("beta", "coverage"),
  nthreads = 1
))

regions_gr <- list(run_test(
  biscuit_bedfiles,
  gr,
  funcs = "all",
  c(4, 5),
  c("beta", "coverage"),
  nthreads = 1
))


full_set <- list(
  one_file,
  one_fun,
  two_fun,
  one_reg,
  threads,
  regions_df,
  regions_gr
)

test_dims <- function(set) {
  lapply(set, function(i) {
    expect_equal(
      i$dims,
      dim(i$df)
    )
  })
}

test_colnames <- function(set) {
  lapply(set, function(i) {
    expect_equal(
      i$colnames,
      colnames(i$df)
    )
  })
}

test_that("dims", {
  lapply(full_set[6], function(i) {
    test_dims(i)
  })
})

test_that("colnames", {
  lapply(full_set, function(i) {
    test_colnames(i)
  })
})

# error on bad 'fun' arguments
test_that("bad fun", {
  expect_error(
    summarize_regions(biscuit_bedfiles, regions, fun = "none")
  )
  expect_error(
    summarize_regions(biscuit_bedfiles, regions, fun = c("all", "mean"))
  )
  expect_error(
    summarize_regions(biscuit_bedfiles, regions, fun = c("none", "mean"))
  )
})

#' Set and get logging level
#'
#' @param level The logging verbosity level to use
#' - `"info"`: the default that gives provides basic information about the
#' number of files and regions used in a function
#' - `"debug"`: more verbose about row allocations, how many CpGs were found in
#' a region, filename parsing etc. This mode cannot be used on more than one
#' thread as R cannot output messages from multiple threads without crashing.
#' - `"off"`: no logging
#' @returns
#'  - `set_log_level()`: None; sets the log level to the provided level
#'  - `get_log_level()`: The current logging level as a string
#'
#' @export
#' @order 1
#'
#' @examples
#' set_log_level("info")
#' get_log_level()
set_log_level <- function(level = "info") {
  n_threads <- get_threads()["use_threads"]
  validate_log_level(level, n_threads)
  Cpp_set_log_level(level)
}

#' Validate provided logging level
#'
#' Only "info" and "debug" are currently supported, with "debug" only supported
#' when using 1 thread
#' @param level The logging level to validate
#' @param n_threads The number of threads that the next iscream function call
#' will use
#' @returns None; sets the log level to the provide level
#'
#' @keywords internal
#'
#' @examples
#' set_log_level("info")
validate_log_level <- function(level = get_log_level(), n_threads) {
  quiet_logging <- c("error", "info", "off")
  supported <- c(quiet_logging, "trace", "debug", "warn")
  unsupported <- c("critical")

  if (!(level %in% supported)) {
    if (level %in% unsupported) {
      stop(level, " logging is currently not supported")
    } else {
      stop(level, " is not a supported spdlog log level")
    }
  }

  if (n_threads > 1 && !(level %in% quiet_logging)) {
    stop(
      n_threads,
      " threads were requested ",
      " but ",
      level,
      " logging cannot use more than 1 thread.",
      " Run `set_log_level(\"info\")` to use multiple threads.",
      " See `?set_log_level` for more details."
    )
  }
}

options("iscream.threads" = 1)
test_that("set_log_level to trace on single thread", {
  expect_no_error(set_log_level("trace"))
  expect_equal(get_log_level(), "trace")
})

test_that("set_log_level to debug on single thread", {
  expect_no_error(set_log_level("debug"))
  expect_equal(get_log_level(), "debug")
})

test_that("set_log_level to info on single thread", {
  expect_no_error(set_log_level("info"))
  expect_equal(get_log_level(), "info")
})

test_that("set_log_level to debug on single thread", {
  expect_no_error(set_log_level("warn"))
  expect_equal(get_log_level(), "warn")
})

test_that("set_log_level to debug on single thread", {
  expect_no_error(set_log_level("error"))
  expect_equal(get_log_level(), "error")
})

options("iscream.threads" = 2)

test_that("set_log_level to error on 2 threads", {
  expect_no_error(set_log_level("error"))
  expect_equal(get_log_level(), "error")
})

test_that("set_log_level to info on 2 threads", {
  expect_no_error(set_log_level("info"))
  expect_equal(get_log_level(), "info")
})

test_that("set_log_level to info on 2 threads", {
  expect_no_error(set_log_level("off"))
  expect_equal(get_log_level(), "off")
})

test_that("set_log_level to warn on 2 threads", {
  expect_error(set_log_level("warn"))
})

test_that("set_log_level to debug on 2 threads", {
  expect_error(set_log_level("debug"))
})

test_that("set_log_level to warn on 2 threads", {
  expect_error(set_log_level("warn"))
})

test_that("unsupported log level", {
  expect_error(set_log_level("warn"))
})

test_that("invalid log level", {
  expect_error(set_log_level("invalid"))
})

test_that("unsupported log level", {
  expect_error(set_log_level("critical"))
})

#' Validate provided aligner
#'
#' Only "biscuit", "bismark", and "bsbolt" are currently supported
#' @param aligner The input alinger
#' @returns true; quits if the input is not among supported_aligners
#'
#' @keywords internal
verify_aligner_or_stop <- function(aligner) {
  supported_aligners <- c("biscuit", "bismark", "bsbolt")
  if (!(aligner %in% supported_aligners)) {
    stop(paste(
      "aligner =",
      aligner,
      "not supported. Use one of the supported aligners:",
      paste(supported_aligners, collapse = ", ")
    ))
  }
}

#' GRanges to region strings
#'
#' Coerces GenomicRanges to `chr:start-end` strings with `as.character`. If any
#' regions have the same start and end, `as.character` returns `chr:start`
#' strings which are invalid for the htslib API. These are corrected to
#' `chr:start-start`.
#'
#' @param gr A GRanges object
#' @returns A character vector
#'
#' @export
#' @examples
#' if (requireNamespace("GenomicRanges", quietly = TRUE)) {
#'   get_granges_string(GenomicRanges::GRanges(c("chr1:1-10", "chr2:15-20")))
#' }
get_granges_string <- function(gr) {
  if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
    stop("The 'GenomicRanges' package must be installed for this functionality")
  }
  region_str <- as.character(gr)
  singles <- which(GenomicRanges::width(gr) == 1)
  single_count <- length(singles)
  if (single_count > 0) {
    region_str[singles] <- gsub(":([0-9]+)", ":\\1-\\1", region_str[singles])
    message("Corrected ", single_count, " invalid 'chr:start' region strings to 'chr:start-start'")
  }

  names(region_str) <- names(gr)
  region_str
}

#' DataFrame to region strings
#'
#' Convert DataFrame to a vector of strings. Set feature names in a "name" column
#'
#' @param regions_df A data frame with "chr", "start" and "end" columns
#' @param feature_col The data frame column to use as the names of the output string vector
#'
#' @importFrom data.table setDT
#'
#' @returns A character vector
#'
#' @export
#' @examples
#' (df <- data.frame(chr = c("chr1", "chr2"), start = c(1, 5), end = c(4, 10)))
#' get_df_string(df)
get_df_string <- function(regions_df, feature_col = NULL) {
  colnames.check <- colnames(regions_df)[seq_len(3)]
  stopifnot(
    "colnames must be 'chr', 'start' and 'end'" = colnames.check == c("chr", "start", "end")
  )
  chr <- start <- end <- NULL
  regions.dt <- setDT(regions_df)
  regions <- regions.dt[, paste0(chr, ":", start, "-", end)]
  if (!is.null(feature_col) && feature_col %in% colnames(regions_df)) {
    names(regions) <- regions_df[[feature_col]]
  }
  return(regions)
}

get_df_from_string <- function(regions) {
  start <- NULL
  as.data.table(regions)[, tstrsplit(regions, ":|-", fixed = FALSE, names = c("chr", "start", "end"))][,
    start := as.integer(start)
  ]
}

# Get GRanges from chr and pos vector
getGR <- function(chr, pos) {
  if (requireNamespace("GenomicRanges", quietly = TRUE)) {
    GenomicRanges::GRanges(chr, IRanges::IRanges(pos, pos, width = 1))
  }
}

# Check if package is loaded
# https://github.com/HenrikBengtsson/R.utils/blob/74def095eaa244e355d05fdf790ee6393dad1d99/R/isPackageLoaded.R#L33-L43
is_package_loaded <- function(package, caller, fail) {
  loaded_packages <- gsub("package:", "", search())
  if (package %in% loaded_packages) {
    return(TRUE)
  }
  if (fail) {
    msg <- sprintf("Please load the '%s' package to use '%s' output", package, caller)
    stop(msg)
  }
  FALSE
}

bismark1 <- paste0(letters[1:10], ".cov.gz")
bismark2 <- paste0(letters[1:10], ".cov")
biscuit1 <- paste0(letters[1:10], ".bed.gz")
biscuit2 <- paste0(letters[1:10], ".bed")
biscuit3 <- paste0(letters[1:10], ".mergecg.bed.gz")
bismark <- c(bismark1, bismark2)
biscuit <- c(biscuit1, biscuit2, biscuit3)

test_that("verify_filetype valid bismark", {
  expect_no_error(iscream:::verify_filetype(bismark, "bismark"))
})

test_that("verify_filetype valid biscuit", {
  expect_no_error(iscream:::verify_filetype(bismark, "bsbolt"))
})

test_that("verify_filetype valid", {
  expect_no_error(iscream:::verify_filetype(biscuit, "biscuit"))
})

test_that("verify_filetype warning", {
  expect_warning(
    iscream:::verify_filetype(biscuit, "bismark"),
    "'aligner' set to bismark but no files found with '.cov', extension - verify aligner and data frame colnames"
  )
  expect_warning(
    iscream:::verify_filetype(biscuit, "bsbolt"),
    "'aligner' set to bsbolt but no files found with '.cov', extension - verify aligner and data frame colnames"
  )
  expect_warning(
    iscream:::verify_filetype(bismark, "biscuit"),
    "'aligner' set to 'biscuit' but files found with '.cov', extension - verify aligner and data frame colnames"
  )
})

test_that("verify_filetype error", {
  expect_error(
    iscream:::verify_filetype(biscuit, "bismark", stop_on_error = TRUE),
    "'aligner' set to bismark but no files found with '.cov', extension - verify aligner"
  )
  expect_error(
    iscream:::verify_filetype(biscuit, "bsbolt", stop_on_error = TRUE),
    "'aligner' set to bsbolt but no files found with '.cov', extension - verify aligner"
  )
  expect_error(
    iscream:::verify_filetype(bismark, "biscuit", stop_on_error = TRUE),
    "'aligner' set to 'biscuit' but files found with '.cov', extension - verify aligner"
  )
})

test_that("verify_aligner valid", {
  expect_no_error(iscream:::verify_aligner_or_stop("biscuit"))
  expect_no_error(iscream:::verify_aligner_or_stop("bismark"))
  expect_no_error(iscream:::verify_aligner_or_stop("bsbolt"))
})

test_that("verify_aligner_or_stop invalid", {
  expect_error(iscream:::verify_aligner_or_stop("unsupported"))
})

test_that("is_package_loaded", {
  expect_true(iscream:::is_package_loaded("base", "NULL", fail = FALSE))
  expect_false(iscream:::is_package_loaded("does_not_exist", "NULL", fail = FALSE))
  expect_error(
    iscream:::is_package_loaded("does_not_exist", "tester", fail = TRUE),
    "Please load the 'does_not_exist' package to use 'tester' output"
  )
})

options("iscream.threads" = 1)
regions <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14")
gr <- GenomicRanges::GRanges(regions)

regions_single <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14", D = "chr2:5", E = "chrX:3")
regions_single_res <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14", D = "chr2:5-5", E = "chrX:3-3")
gr_single <- GenomicRanges::GRanges(regions_single)

with_mocked_bindings(
  {
    expect_error(
      get_granges_string(gr),
      "The 'GenomicRanges' package must be installed for this functionality"
    )
  },
  requireNamespace = function(package, ..., quietly = FALSE) FALSE,
  .package = "base"
)

test_that("Test GRanges to string", {
  str <- get_granges_string(gr)
  expect_equal(str, regions)
  expect_equal(names(str), names(regions))
  str <- get_granges_string(gr_single)
  expect_equal(str, regions_single_res)
  expect_equal(names(str), names(regions_single))
})

library(data.table)

regions.dt <- as.data.table(regions, keep.rownames = TRUE)[, tstrsplit(regions, ":|-")]

test_that("Test GRanges to string", {
  expect_error(get_df_string(regions.dt))
})

colnames(regions.dt) <- c("chr", "start", "end")
test_that("Test GRanges to string no names", {
  expect_equal(get_df_string(regions.dt), unname(regions))
})

regions.dt[, names := names(regions)]
colnames(regions.dt) <- c("chr", "start", "end", "names")
test_that("Test GRanges to string with names", {
  expect_equal(get_df_string(regions.dt, feature_col = "names"), regions)
  expect_equal(names(get_df_string(regions.dt, feature_col = "names")), names(regions))
})

#' Get the number of available threads
#'
#' Gets the number of threads iscream is currently set to use, whether the
#' `"iscream.threads"` option is set and how many threads are available for
#' use. To set the number of threads use `set_threads()` or set the
#' `iscream.threads` option in your `~/.Rprofile`. See `?set_threads` for more
#' information.
#' @importFrom parallelly availableCores
#' @returns A named vector:
#' - `use_threads` = the number of threads iscream will use
#' - `opt_set` = whether the option was set by the user
#' - `avail_threads` = The number of available threads as reported by
#' `parallelly::availableCores`
#'
#' @export
#'
#' @examples
#' get_threads()
get_threads <- function() {
  opt_threads <- unname(getOption("iscream.threads"))
  avail_threads <- unname(availableCores())
  if (is.null(opt_threads)) {
    return(c("use_threads" = 1, "opt_set" = FALSE, "avail_threads" = avail_threads))
  }
  check_thread_count(opt_threads, avail_threads, TRUE)
  c("use_threads" = opt_threads, "opt_set" = TRUE, "avail_threads" = avail_threads)
}

# Get the number of threads to use in functions from param or option
.get_threads <- function(nthreads) {
  ifelse(
    is.null(nthreads),
    getOption("iscream.threads"),
    check_thread_count(nthreads)
  )
}

#' Set the number of available threads
#'
#' Sets the `"iscream.threads"` option to `n_threads`. To see how many threads
#' you have available see `?get_threads()`.

#' @param n_threads The number of threads to use
#' @importFrom parallelly availableCores
#' @returns None. Sets the `iscream.threads` option to the requested number of
#' threads if available
#'
#' @details iscream uses OpenMP to parallelize certain functions. You can use
#' as many threads as are available to you on your system to varying degrees of
#' performance improvements. The `get_threads()` function uses
#' `parallelly::availableCores()` to report the number of available threads.
#' Although OpenMP can detect the number of available cores, on high
#' performance computers (HPCs) with resource allocating job schedulers like
#' SLURM, OpenMP may detect all available threads across the HPC and not limit
#' itself to the cores that were allocated to you by the scheduler. If your
#' system administrator has not set up any limits, this may result in your job
#' taking resources from other jobs. If there are limits, trying to use more
#' threads that those available will reduce iscream's performance. Job
#' schedulers will typically have an environment variable (e.g.
#' `SLURM_CPUS_ON_NODE` with SLURM) that gives you the actual number of
#' available cores. Further, on hyperthreaded systems, this count may be double
#' that of the available processors. Using hyperthreading does not guarantee
#' any performance improvement - it may be better to set the number of threads
#' to half the reported number. `parallelly::availableCores()` takes HPC
#' scheduler/CRAN/Bioconductor limits into account when reporting the number of
#' available threads but it may not reliably report hyperthreading ('system' or
#' 'nproc'). To set the number of threads without having to call
#' `set_threads()` in every session, put
#'
#' ```
#' options(iscream.threads = [n_threads])
#' ```
#' in your `.Rprofile` See `help('Rprofile')` for information on startup options.
#'
#' Functions currently using multithreading:
#' - `tabix()`, `tabix_gr()`, `tabix_raw()`
#' - `query_chroms()`
#' - `make_mat()`, `make_mat_se()`, `make_mat_gr()`, `make_mat_bsseq()`
#' - `summarize_regions()`, `summarize_meth_regions()`
#'
#' @export
#'
#' @examples
#' (ncores <- parallelly::availableCores())
#' \donttest{
#' set_threads(ncores)
#' }
set_threads <- function(n_threads) {
  check_thread_count(n_threads)
  options(iscream.threads = n_threads)
  message("iscream now using ", n_threads, " of ", availableCores(), " available threads.")
}

#' Check that the required threads are available
#'
#' @param n_threads The number of threads to check availability for
#' @param avail_threads The number of threads that are available on the system.
#' Defaults to `parallelly::availableCores()`
#' @param opt_set Whether the `iscream.threads` options is set
#'
#' @importFrom parallelly availableCores
#' @returns `n_threads` if the requested number of threads are available and
#' stops if not
#'
#' @keywords internal
check_thread_count <- function(
  n_threads,
  avail_threads = availableCores(),
  opt_set = FALSE
) {
  if (n_threads <= avail_threads) {
    return(n_threads)
  }
  if (opt_set) {
    msg <- paste0("The `iscream.threads` option is set to ", n_threads, " threads but your")
  } else {
    msg <- paste0("Cannot use ", n_threads, " threads. Your")
  }
  stop(
    msg,
    " system has only ",
    avail_threads,
    " threads. See parallelly::availableCores(which = 'all') for more informaion on available resources"
  )
}

options("iscream.threads" = 2)
library("parallelly")
max_threads <- 2

test_that("set_threads within limit", {
  set_threads(1)
  expect_equal(getOption("iscream.threads"), 1)
  set_threads(max_threads)
  expect_equal(getOption("iscream.threads"), max_threads)
})

test_that("set_threads over limit", {
  expect_error(set_threads(999))
  expect_error(set_threads(availableCores() + 1))
})

test_that("set threads option over limit", {
  expect_error(check_thread_count(999, opt_set = TRUE))
})

options("iscream.threads" = availableCores())
test_that("test thread setting onAttach", {
  expect_no_error(iscream:::package_loader())
})

options("iscream.threads" = NULL)
expected_threads <- c(
  'use_threads' = 1,
  opt_set = FALSE,
  avail_threads = unname(availableCores())
)

test_that("get threads with NULL option", {
  expect_equal(get_threads(), expected_threads)
})

#' Summarize methylation information over genomic regions
#'
#' Run summarizing functions on the CpG/CpH loci in BED files across genomic
#' regions. Parallelized across files using threads from the
#' `"iscream.threads"` option.
#'
#' @inheritParams summarize_regions
#'
#' @param aligner The aligner used to produce the BED files - one of "biscuit",
#' "bismark", "bsbolt".
#' @param mval Whether to calculate the M value (coverage \eqn{\times \beta})
#' or use the beta value when applying the function.
#'
#' @inheritSection summarize_regions Supported functions
#'
#' @inheritSection summarize_regions Using feature identifiers
#'
#' @importFrom methods is
#'
#' @returns A data.frame
#'
#' @export
#'
#' @examples
#' # also see examples from ?summarize_regions
#'
#' bedfiles <- system.file("extdata", package = "iscream") |>
#'   list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
#'
#' # make a vector of regions
#' regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
#' summarize_meth_regions(bedfiles, regions)
#' names(regions) <- c("A", "B", "C")
#' summarize_meth_regions(bedfiles, regions, fun = c("mean", "stddev"), mval = FALSE)
#' summarize_meth_regions(bedfiles, regions, fun = "sum")
summarize_meth_regions <- function(
  bedfiles,
  regions,
  fun = "all",
  aligner = "biscuit",
  feature_col = NULL,
  mval = TRUE,
  set_region_rownames = FALSE,
  nthreads = NULL
) {
  supported_funcs <- c("sum", "mean", "median", "stddev", "variance", "min", "max", "range", "count")

  if (length(fun) > 1) {
    if ("all" %in% fun) {
      stop("'all' can't be used with other summary funcions")
    }
    stopifnot("Selected function not supported" = all(fun %in% supported_funcs))
    fun_to_use <- fun
  } else {
    stopifnot("Selected function not supported" = fun %in% c(supported_funcs, "all"))
    fun_to_use <- supported_funcs
    if (fun != "all") {
      fun_to_use <- fun
    }
  }

  if (aligner != "general") {
    col_names <- c("coverage", ifelse(mval, "M", "beta"))
  }

  stopifnot("'mval' must be TRUE or FALSE" = mval %in% c(TRUE, FALSE))

  verify_aligner_or_stop(aligner)
  verify_files_or_stop(bedfiles, verify_tabix = TRUE)
  verify_filetype(bedfiles, aligner, stop_on_error = TRUE)
  if (is(regions, "GRanges")) {
    regions <- get_granges_string(regions)
  } else if ("data.frame" %in% class(regions)) {
    regions <- get_df_string(regions, feature_col)
  }

  n_threads <- .get_threads(nthreads)
  validate_log_level(n_threads = n_threads)

  df <- Cpp_summarize_regions(
    bedfiles = bedfiles,
    regions = regions,
    col_indices = c(4, 5),
    col_names = col_names,
    fun_vec = fun_to_use,
    aligner = aligner,
    mval = mval,
    region_rownames = set_region_rownames,
    nthreads = n_threads
  )
  df[df == -99] <- NA

  mval_count <- paste0(ifelse(mval, "M", "beta"), ".count")
  if (mval_count %in% colnames(df)) {
    df <- df[, !(names(df) %in% "coverage.count")]
  }

  colnames(df)[which(colnames(df) == mval_count)] <- "cpg_count"
  df
}

library("stringfish")
library("data.table")

# inputs
extdata <- system.file("extdata", package = "iscream")
biscuit_bedfiles <- list.files(extdata, pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
bismark_bedfiles <- list.files(extdata, pattern = "[a|b|c|d].cov.gz$", full.names = TRUE)
regions <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14")
regions.dt <- as.data.table(regions)[, tstrsplit(regions, ":|-")][, names := names(regions)]
colnames(regions.dt) <- c("chr", "start", "end", "names")
gr <- GenomicRanges::GRanges(regions)
supported_funcs <- c("sum", "mean", "median", "stddev", "variance", "min", "max", "range")

# utils
get_meth_colnames <- function(mval, funcs) {
  base_colnames <- c("feature", "file")
  values <- c("coverage", ifelse(mval, "M", "beta"))
  if ("all" %in% funcs) {
    return(c(base_colnames, as.vector(outer(values, supported_funcs, paste, sep = ".")), "cpg_count"))
  }

  if ("count" %in% funcs) {
    index <- which(funcs == "cpg_count")
    funcs <- funcs[-index]
  }
  function_used <- funcs
  col_names <- c(
    base_colnames,
    as.vector(outer(values, function_used, paste, sep = "."))
  )

  if ("cpg_count" %in% funcs) {
  col_names <- c(col_names, "cpg_count")
  }

  return(col_names)
}

run_meth_test <- function(bedfiles, regions, funcs, mval, nthreads) {
  colnames <- get_meth_colnames(mval, funcs)
  list(
    df = summarize_meth_regions(bedfiles, regions, fun = funcs, mval = mval, nthreads = nthreads),
    colnames = colnames,
    dims = c(length(bedfiles) * length(regions), length(colnames))
  )
}

one_file <- lapply(biscuit_bedfiles, function(i) {
  run_meth_test(
    i,
    regions,
    funcs = "all",
    mval = FALSE,
    nthreads = 1
  )
})

one_fun <- lapply(supported_funcs, function(i) {
  run_meth_test(
    biscuit_bedfiles,
    regions,
    funcs = i,
    mval = FALSE,
    nthreads = 1
  )
})

two_fun <- lapply(asplit(combn(supported_funcs, 2), 2), function(i) {
  run_meth_test(
    biscuit_bedfiles,
    regions,
    funcs = i,
    mval = FALSE,
    nthreads = 1
  )
})

one_reg <- lapply(regions, function(i) {
  run_meth_test(
    biscuit_bedfiles,
    i,
    funcs = "all",
    mval = FALSE,
    nthreads = 1
  )
})

mbeta <- lapply(c(T, F), function(i) {
  run_meth_test(
    biscuit_bedfiles,
    regions,
    funcs = "all",
    mval = i,
    nthreads = 1
  )
})

threads <- lapply(c(1, 1), function(i) {
  run_meth_test(
    biscuit_bedfiles,
    regions,
    funcs = "all",
    mval = F,
    nthreads = i
  )
})

full_set <- list(
  one_file,
  one_fun,
  two_fun,
  one_reg,
  mbeta,
  threads
)

test_dims <- function(set) {
  lapply(set, function(i) {
    expect_equal(
      i$dims,
      dim(i$df)
    )
  })
}

test_colnames <- function(set) {
  lapply(set, function(i) {
    expect_equal(
      i$colnames,
      colnames(i$df)
    )
  })
}

test_that("dims", {
  lapply(full_set, function(i) {
    test_dims(i)
  })
})

test_that("colnames", {
  lapply(full_set, function(i) {
    test_colnames(i)
  })
})

# error on bad 'fun' arguments
test_that("bad fun", {
  expect_error(
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "none")
  )
  expect_error(
    summarize_meth_regions(biscuit_bedfiles, regions, fun = c("all", "mean"))
  )
  expect_error(
    summarize_meth_regions(biscuit_bedfiles, regions, fun = c("none", "mean"))
  )
})

m_sum <- read.csv(file.path(extdata, "summarize_regions_m_sum.test"))
beta_sum <- read.csv(file.path(extdata, "summarize_regions_beta_sum.test"))
m_mean <- read.csv(file.path(extdata, "summarize_regions_m_mean.test"))
beta_mean <- read.csv(file.path(extdata, "summarize_regions_beta_mean.test"))
beta_all <- read.csv(file.path(extdata, "summarize_regions_beta_all.test"))
m_all <- read.csv(file.path(extdata, "summarize_regions_m_all.test"))

test_that("summarize_meth_regions 1 thread sum", {
  expect_equal(
    m_sum,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "sum", mval = TRUE, nthreads = 1)
  )
  expect_equal(
    m_sum,
    summarize_meth_regions(biscuit_bedfiles, gr, fun = "sum", mval = TRUE, nthreads = 1)
  )
  expect_equal(
    m_sum,
    summarize_meth_regions(biscuit_bedfiles, regions.dt, fun = "sum", mval = TRUE, feature_col = "names", nthreads = 1)
  )
  expect_equal(
    beta_sum,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "sum", mval = FALSE, nthreads = 1)
  )
})

test_that("summarize_meth_regions 2 thread sum", {
  expect_equal(
    m_sum,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "sum", mval = TRUE, nthreads = 2)
  )
  expect_equal(
    beta_sum,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "sum", mval = FALSE, nthreads = 2)
  )
})

test_that("summarize_meth_regions 1 thread sum", {
  expect_equal(
    m_sum,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "sum", mval = TRUE, nthreads = 1)
  )
  expect_equal(
    beta_sum,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "sum", mval = FALSE, nthreads = 1)
  )
})

test_that("summarize_meth_regions 2 thread sum", {
  expect_equal(
    m_sum,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "sum", mval = TRUE, nthreads = 2)
  )
  expect_equal(
    beta_sum,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "sum", mval = FALSE, nthreads = 2)
  )
})


test_that("summarize_meth_regions 1 thread all", {
  expect_equal(
    m_all,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "all", mval = TRUE, nthreads = 1)
  )
  expect_equal(
    beta_all,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "all", mval = FALSE, nthreads = 1)
  )
})

test_that("summarize_meth_regions 2 thread all", {
  expect_equal(
    m_all,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "all", mval = TRUE, nthreads = 2)
  )
  expect_equal(
    beta_all,
    summarize_meth_regions(biscuit_bedfiles, regions, fun = "all", mval = FALSE, nthreads = 2)
  )
})


test_that("summarize_meth_regions 2 thread all bismark", {
  expect_equal(
    m_all,
    summarize_meth_regions(bismark_bedfiles, regions, fun = "all", mval = TRUE, aligner = "bismark", nthreads = 2)
  )
  expect_equal(
    m_all,
    summarize_meth_regions(bismark_bedfiles, gr, fun = "all", mval = TRUE, aligner = "bismark", nthreads = 2)
  )
  expect_equal(
    m_all,
    summarize_meth_regions(bismark_bedfiles, regions.dt, fun = "all", mval = TRUE, aligner = "bismark", feature_col = "names", nthreads = 2)
  )
  expect_equal(
    beta_all,
    summarize_meth_regions(bismark_bedfiles, regions, fun = "all", mval = FALSE, aligner = "bismark", nthreads = 2)
  )
})

test_that("summarize_meth_regions rownames", {
  expect_equal(
    rownames(summarize_meth_regions(biscuit_bedfiles, regions, set_region_rownames = T)),
    rep(unname(regions), length(biscuit_bedfiles))
  )
})

#' Check that files exist
#'
#' @param files_vec A vector of file paths
#' @param error_file_prefix Error message prefix for 'Bedfile' vs 'Tabix file'
#' @returns TRUE if all input BED files have an associated tabix index file.
#' FALSE if not
#'
#' @keywords internal
check_files_exist <- function(files_vec, error_file_prefix = "Bedfile") {
  valid_files <- file.exists(files_vec)
  missing_files <- files_vec[!valid_files]
  if (length(missing_files != 0)) {
    stop(error_file_prefix, ": ", missing_files, " could not be found\n")
  }
}

#' Verify that BED files are tabixed
#'
#' @param bedfiles A vector of BED file paths
#' @param verify_tabix Whether to verify the presence of tabix files
#' @returns TRUE if all input BED files have an associated tabix index file.
#' FALSE if not
#'
#' @keywords internal
verify_files_or_stop <- function(bedfiles, verify_tabix = TRUE) {
  check_files_exist(bedfiles)
  if (verify_tabix) {
    tbi_files <- paste0(bedfiles, ".tbi")
    check_files_exist(tbi_files, "Tabix file")
  }
}

#' Verify that the input BED files are of the type specified by the input aligner
#'
#' @param bedfiles A vector of BED file paths
#' @param aligner The aligner chosen
#' @param stop_on_error Whether to warn or stop on aligner-filename mismatch
#'
#' @importFrom stringfish sf_grepl
#'
#' @returns TRUE if all input BED files have an associated tabix index file.
#' FALSE if not
#'
#' @keywords internal
verify_filetype <- function(bedfiles, aligner, stop_on_error = FALSE) {
  warn_stop <- ifelse(stop_on_error, stop, warning)
  check_name_warning <- "- verify aligner"
  warning_msg <- ifelse(stop_on_error, check_name_warning, paste(check_name_warning, "and data frame colnames"))

  if (aligner == "biscuit" & any(sf_grepl(bedfiles, pattern = ".cov"))) {
    warn_stop("'aligner' set to 'biscuit' but files found with '.cov', extension ", warning_msg)
  }
  if (aligner != "biscuit" & any(!sf_grepl(bedfiles, pattern = ".cov"))) {
    warn_stop("'aligner' set to ", aligner, " but no files found with '.cov', extension ", warning_msg)
  }
}

data_loc <- system.file("extdata", package = "iscream")
bedfiles <- list.files(data_loc, pattern = "*.bed.gz$", full.names = TRUE)
bedfiles_no_path <- list.files(data_loc, pattern = "[a|b].bed.gz$", full.names = FALSE)

test_that("check_files_exist", {
  expect_error(
    iscream:::check_files_exist(bedfiles_no_path[1]),
    paste0("Bedfile: ", bedfiles_no_path[1], " could not be found")
  )
  expect_error(iscream:::verify_files_or_stop(bedfiles_no_path))
  non_tabixed_file <- list.files(data_loc, pattern = "no_tabix.bed.gz", full.names = TRUE)
  expect_error(
    iscream:::verify_files_or_stop(non_tabixed_file),
    paste0("Tabix file: ", non_tabixed_file, ".tbi could not be found")
  )
})

test_that("verify_regions_or_stop invalid", {
  expect_error(iscream:::verify_regions_or_stop(invalid_regions, 1))
})

package_loader <- function() {
  threads <- get_threads()
  use_threads <- threads["use_threads"]
  opt_set <- threads["opt_set"]
  avail_threads <- threads["avail_threads"]
  options("iscream.threads" = unname(use_threads))
  msg <- "iscream using "
  help_msg <- " See `?set_threads` for information on multithreading"

  if (use_threads == avail_threads) {
    msg <- paste0(
      msg,
      use_threads,
      " threads of detected ",
      avail_threads,
      " threads.",
      help_msg,
      "."
    )
  } else {
    msg <- paste0(
      msg,
      use_threads,
      ifelse(use_threads > 1, " threads", " thread"),
      ifelse(opt_set, " based on 'options(iscream.threads)'", " by default"),
      " but parallelly::availableCores() detects ",
      avail_threads,
      " possibly available threads.",
      help_msg,
      " before trying to use more."
    )
  }

  if (Sys.which("tabix") == "") {
    options("tabix.method" = "htslib")
    msg <- paste0(
      msg,
      "\n'tabix' executable not found in $PATH.",
      " tabix() will use htslib to make queries instead which can be slower.",
      " See ?tabix for details."
    )
  } else {
    options("tabix.method" = "shell")
  }

  return(msg)
}

.onAttach <- function(libname, pkgname) {
  msg <- package_loader()
  packageStartupMessage(msg)
  setup_logger()
  invisible()
}

#' Make M/beta and coverage matrices from WGBS BED files
#'
#' Queries the CpG/CpH loci from provided regions and produces M/beta and
#' coverage matrices with their genomic positions. Parallelized across files
#' using threads from the `"iscream.threads"` option. The output of
#' `make_mat_bsseq` may be used to create a BSseq object: `do.call(BSseq,
#' make_mat_bsseq(...))`.
#'
#' @inheritParams make_mat
#'
#' @param aligner The aligner used to produce the BED files - one of "biscuit",
#' "bismark", "bsbolt".
#' @param mval Whether to return M-values or beta-values with the coverage
#' matrix. Defaults to M-value. Set `mval=FALSE` to get beta value matrix.
#' @param merged Whether the input strands have been merged/collapsed
#'
#' @returns A named list of
#' - coverage and either a beta- or M-value matrix
#' - a character vector of chromosomes and numeric vector of corresponding CpG
#' base positions
#' - a character vector of the input sample names
#'
#' @inherit make_mat details
#'
#' @section Bitpacking limits:
#' `make_mat_bsseq()` makes two matrices: M-value (or beta-value) and coverage.
#' For speed and memory efficiency these two values are bitpacked during matrix
#' creation so that only one matrix needs to be populated and resized. This
#' matrix is unpacked into the two required matrices only after the matrix
#' dimensions are known after querying all input files. The two values are
#' packed using the INT16 type, which has an upper limit of 32,767, into one
#' INT32. If the coverage values exceed 32,767, the upper limit of a 16-bit
#' signed integer, it will be capped at the limit. Beta values will also be
#' capped similarly, but any such beta values would indicate a bug in the
#' aligner that produced the data.
#'
#' @importFrom Matrix drop0
#' @importFrom methods is
#'
#' @export
#' @examples
#' bedfiles <- system.file("extdata", package = "iscream") |>
#'   list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
#' # examine the BED files
#' colnames <- c("chr", "start", "end", "beta", "coverage")
#' lapply(bedfiles, function(i) knitr::kable(read.table(i, col.names = colnames)))
#'
#' # make a vector of regions
#' regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
#' mat <- make_mat_bsseq(bedfiles, regions)
#' # for BSseq object run
#' if (requireNamespace("bsseq", quietly = TRUE)) {
#'   do.call(bsseq::BSseq, mat)
#' }
make_mat_bsseq <- function(
  bedfiles,
  regions,
  aligner = "biscuit",
  mval = TRUE,
  merged = TRUE,
  sparse = FALSE,
  prealloc = 10000,
  nthreads = NULL
) {
  verify_aligner_or_stop(aligner)
  verify_filetype(bedfiles, aligner, stop_on_error = TRUE)
  verify_files_or_stop(bedfiles, verify_tabix = TRUE)
  if (is(regions, "GRanges")) {
    regions <- get_granges_string(regions)
  } else if ("data.frame" %in% class(regions)) {
    regions <- get_df_string(regions)
  }

  n_threads <- .get_threads(nthreads)
  validate_log_level(n_threads = n_threads)

  b <- Cpp_query_all(
    bedfiles = bedfiles,
    regions = regions,
    aligner = aligner,
    valInd = 0,
    merged = merged,
    sparse = sparse,
    prealloc = prealloc,
    nthreads = n_threads
  )

  if (sparse) {
    get_beta_m <- ifelse(mval, get_m_sparse, get_beta_sparse)
  } else {
    get_beta_m <- ifelse(mval, get_m, get_beta)
  }

  if (sparse) {
    get_beta_m(b$M)
    b$M <- drop0(b$M)
    get_cov_sparse(b$Cov)
  } else {
    get_beta_m(b$M, n_threads)
    get_cov(b$Cov, n_threads)
  }
  if (!mval) {
    names(b)[which(names(b) == "M")] <- "beta"
  }
  b
}

#' Query the chromosomes or seqnames from a vector of BED files
#' @param bedfiles The vector of BED file paths
#' @param nthreads Set number of threads to use overriding the
#' `"iscream.threads"` option. See `?set_threads` for more information.
#' @returns A vector of seqnames
#'
#' @export
#' @examples
#' bedfiles <- system.file("extdata", package = "iscream") |>
#'   list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
#' query_chroms(bedfiles)
query_chroms <- function(bedfiles, nthreads = NULL) {
  verify_files_or_stop(bedfiles)

  n_threads <- .get_threads(nthreads)

  Cpp_query_chroms(bedfiles, n_threads)
}

#' Query records from tabixed BED files
#' @param bedfiles The BED files to be queried
#' @param regions A vector, data frame or GenomicRanges of genomic regions. See
#' details.
#' @param aligner The aligner used to produce the BED files - one of "biscuit",
#' "bismark", "bsbolt". Will set the result data.table's column names based on
#' this argument.
#' @param col.names A vector of column names for the data columns of the
#' result.table, not including "chr", "start", and "end". Set if your BED file
#' is not from the supported aligners or is a general BED file.
#' @param nthreads Set number of threads to use overriding the
#' `"iscream.threads"` option. See `?set_threads` for more information.
#'
#' @details
#'
#' ## Query method
#' '*iscream* has two methods to query records from BED files:
#' - the *tabix* shell executable: fast since its output can be redirected to a
#' file (which `data.table::fread()` can then read) instead of having to
#' allocate memory and store it during the query
#'
#' - *iscream's* tabix implementation, based on the *tabix* executable using
#' *htslib*, but slower on large queries since it stores the records as they
#' are found instead of writing to a file. However it's able to store each
#' region's records independently instead of in a single file and is used in
#' `make_mat()`, `make_mat_bsseq()`, and `summarize_regions()`.
#'
#' When *iscream* is attached, it checks that the *tabix* executable is
#' available with `Sys.which()` and, if available, sets `options("tabix.method"
#' = "shell")`. `tabix()` then uses the *tabix* executable to make queries,
#' except for `tabix_raw()`. If *tabix* is not found, *iscream* uses its tabix
#' implementation. To use only *iscream's* tabix implementation, set
#' `options("tabix.method" = "htslib")`.
#'
#' ## Input region formats
#' The input regions format may be string vector in the form "chr:start-end",
#' a dataframe with "chr", "start" and "end" columns or a `GRanges` object.
#' Input regions must be 1-based. When using `"htslib"` as the query method, if
#' the input `GRanges` object of regions contains any single locus regions
#' where the start and end positions are the same, iscream will notify that
#' such regions were found and fixed as `chr:start` format strings are invalid
#' for the htslib API (see `?get_granges_string`).
#'
#' @importFrom data.table as.data.table tstrsplit set := rbindlist fread fwrite setnames
#' @importFrom pbapply pblapply pboptions
#' @importFrom tools file_path_sans_ext
#' @importFrom stats setNames
#' @importFrom methods is
#' @returns
#' - `tabix()`: A data frame
#' - `tabix_gr()`: A `GRanges` object for single files and `GRangesList` for
#' multiple files. When making `GRanges`, the 0-based records from BED-files
#' will be converted to 1-based with
#' `GenomicRanges::makeGRangesFromDataFrame()`. Bismark's coverage files will
#' not be converted as they are already 1-based and the `ranges` slot will be
#' only one position.
#' - `tabix_raw()`: A named list of raw strings from the regions in the style
#' of `Rsamtools::scanTabix`

#'
#' @export
#' @examples
#' bedfiles <- system.file("extdata", package = "iscream") |>
#'   list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
#' regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
#' tabix(bedfiles, regions, col.names = c("beta", "coverage"))
#' if (require("GenomicRanges", quietly = TRUE)) {
#'   tabix_gr(bedfiles, regions, col.names = c("beta", "coverage"))
#' }
#' tabix_raw(bedfiles, regions)
tabix <- function(
  bedfiles,
  regions,
  aligner = NULL,
  col.names = NULL,
  nthreads = NULL
) {
  verify_files_or_stop(bedfiles)
  if (!is.null(aligner)) {
    verify_aligner_or_stop(aligner)
    verify_filetype(bedfiles, aligner)
  }

  # make the query
  pboptions(char = "*", min_time = 2)
  if (getOption("tabix.method") == "htslib") {
    input_regions <- get_string_input_regions(regions)
    result <- tabix.htslib(bedfiles, input_regions, nthreads)
  } else {
    regions_df <- get_df_input_regions(regions)
    result <- tabix.shell(bedfiles, regions_df, nthreads)
  }

  # single-file empties returns null, multi file returns a non-null empty data.table
  if (is.null(result)) {
    return(NULL)
  } else if (nrow(result) == 0) {
    warning("No records found in any file - if this is unexpected check that your region format matches your bedfiles")
    return(NULL)
  }

  base_colnames <- c("chr", "start", "end")
  if (is.null(aligner)) {
    result_colnames <- get_bed_colnames(col.names, bedfiles, result, base_colnames)
  } else {
    result_colnames <- get_meth_colnames(aligner, bedfiles, result, base_colnames)
  }

  if (length(bedfiles) > 1) {
    result_colnames <- c(result_colnames, "file")
  }

  setnames(result, result_colnames)
  result
}

tabix.shell <- function(bedfiles, regions_df, nthreads) {
  if (length(bedfiles) == 1) {
    return(tabix.shell.single(bedfiles, regions_df))
  }

  pblapply(
    bedfiles,
    function(bedfile) {
      tbx_query <- tabix.shell.single(bedfile, regions_df)
      if (!is.null(tbx_query)) {
        tbx_query[, file := file_path_sans_ext(basename(bedfile), compression = TRUE)]
      }
    },
    cl = .get_threads(nthreads)
  ) |>
    rbindlist()
}

tabix.shell.single <- function(bedfile, regions_df) {
  query.tmpfile <- tempfile(pattern = "regions", fileext = ".tsv")
  if (!is.null(regions_df)) {
    write_bed(regions_df, query.tmpfile)
  }

  cmd <- paste("tabix", bedfile, "-R", query.tmpfile)
  result <- suppressWarnings(fread(cmd = cmd))

  if (is_empty(result, bedfile)) {
    return(NULL)
  }
  result
}

tabix.htslib <- function(bedfiles, input_regions, nthreads) {
  if (length(bedfiles) == 1) {
    return(tabix.htslib.single(bedfile = bedfiles, regions = input_regions))
  }
  dt_list <- pblapply(
    bedfiles,
    function(file) {
      tbx_query <- tabix.htslib.single(
        bedfile = file,
        regions = input_regions
      )
      if (!is.null(tbx_query)) {
        tbx_query[, file := file_path_sans_ext(basename(file), compression = TRUE)]
      }
      return(tbx_query)
    },
    cl = .get_threads(nthreads)
  )
  rbindlist(dt_list)
}

tabix.htslib.single <- function(bedfile, regions) {
  lines <- Cpp_query_interval(bedfile, regions)
  lines_dt <- as.data.table(lines)
  if (is_empty(lines_dt, bedfile)) {
    return(NULL)
  }

  lines_dt[, tstrsplit(lines, "\t", fixed = TRUE, type.convert = TRUE)]
}

run_scan_tabix <- function(bedfiles, input_regions, nthreads) {
  if (length(bedfiles) == 1) {
    return(scan_tabix(bedfiles, input_regions))
  } else {
    bedline_list <- pblapply(bedfiles, scan_tabix, input_regions, cl = .get_threads(nthreads)) |>
      setNames(
        nm = file_path_sans_ext(basename(bedfiles), compression = TRUE),
        object = _
      )
    return(bedline_list)
  }
}

# helpers

get_string_input_regions <- function(regions) {
  if (is(regions, "GRanges")) {
    get_granges_string(regions)
  } else if ("data.frame" %in% class(regions)) {
    get_df_string(regions)
  } else {
    regions
  }
}

get_df_input_regions <- function(regions) {
  if (is(regions, "GRanges")) {
    regions_df <- as.data.table(regions)[, 1:3]
    colnames(regions_df)[1] <- "chr"
    return(regions_df)
  } else if ("data.frame" %in% class(regions)) {
    regions
  } else {
    get_df_from_string(regions)
  }
}

get_meth_colnames <- function(aligner, bedfiles, result, base_colnames) {
  base_colnames <- c("chr", "start", "end")
  biscuit_colnames <- c("beta", "coverage")
  bismark_colnames <- c("methylation.percentage", "count.methylated", "count.unmethylated")

  if (aligner == "biscuit") {
    result_colnames <- c(base_colnames, biscuit_colnames)
    if (grepl("mergecg", bedfiles[1])) {
      result_colnames <- c(result_colnames, "mergecg")
    }
  } else {
    result_colnames <- c(base_colnames, bismark_colnames)
  }
  result_colnames
}

get_bed_colnames <- function(col.names, bedfiles, result, base_colnames) {
  data_cols <- ifelse(length(bedfiles) > 1, ncol(result) - 4, ncol(result) - 3)
  input_length <- length(col.names)

  if (input_length < data_cols) {
    data_colnames <- c(base_colnames, paste0("V", seq_len(data_cols)))
    if (input_length == 0) {
      return(data_colnames)
    }
    warning(
      "Did not use input data 'colnames' - only ",
      length(col.names),
      " names provided for ",
      data_cols,
      " data columns"
    )
    return(data_colnames)
  } else if (length(col.names) > data_cols) {
    warning("Fewer columns in data than provided colnames")
  }

  c(base_colnames, col.names[seq_len(data_cols)])
}


write_bed <- function(regions_df, outfile) {
  fwrite(
    regions_df[, 1:3],
    outfile,
    sep = "\t",
    quote = FALSE,
    col.names = FALSE
  )
}

is_empty <- function(result, bedfile) {
  if (nrow(result) == 0) {
    warning(
      "No records found in ",
      bedfile,
      " - if this is unexpected check that your region format matches your bedfiles"
    )
    return(TRUE)
  }
  return(FALSE)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

decode_beta <- function(encoded) {
    .Call(`_iscream_decode_beta`, encoded)
}

decode_cov <- function(encoded) {
    .Call(`_iscream_decode_cov`, encoded)
}

get_cov <- function(m, nthreads) {
    invisible(.Call(`_iscream_get_cov`, m, nthreads))
}

get_m <- function(m, nthreads) {
    invisible(.Call(`_iscream_get_m`, m, nthreads))
}

get_beta <- function(m, nthreads) {
    invisible(.Call(`_iscream_get_beta`, m, nthreads))
}

get_cov_sparse <- function(m) {
    invisible(.Call(`_iscream_get_cov_sparse`, m))
}

get_m_sparse <- function(m) {
    invisible(.Call(`_iscream_get_m_sparse`, m))
}

get_beta_sparse <- function(m) {
    invisible(.Call(`_iscream_get_beta_sparse`, m))
}

get_omp_threads <- function(verbose) {
    .Call(`_iscream_get_omp_threads`, verbose)
}

setup_logger <- function(logname = "iscream") {
    invisible(.Call(`_iscream_setup_logger`, logname))
}

#' spdlog Logging Lever Setter
#'
#' A helper function to turn a logging level given as string
#' into the current logging level
#'
#' @param name A string with the logging level. Value understood are,
#' in decreasing verbosity \sQuote{trace}, \sQuote{debug}, \sQuote{info},
#' \sQuote{warning}, \sQuote{error}, \sQuote{critical}, and \sQuote{off}.
#' Unrecognised names are equivalent to \sQuote{off}.
#' @return Nothing is returned.
#' @keywords internal
Cpp_set_log_level <- function(name) {
    invisible(.Call(`_iscream_Cpp_set_log_level`, name))
}

#' @rdname set_log_level
#' @order 2
#' @export
get_log_level <- function() {
    .Call(`_iscream_get_log_level`)
}

#' Get htslib version and available features
#'
#' Returns the version of htslib being used by iscream and whether features
#' such as libdeflate support are available. This information may not always
#' correspond to the htslib version used during iscream's installation if a
#' different htslib version is available for linking at runtime.
#'
#' @returns None
#'
#' @examples
#' htslib_version()
#'
#' @export
htslib_version <- function() {
    invisible(.Call(`_iscream_htslib_version`))
}

Cpp_query_chroms <- function(bedfile_vec, nthreads) {
    .Call(`_iscream_Cpp_query_chroms`, bedfile_vec, nthreads)
}

Cpp_query_interval <- function(bedfile, regions) {
    .Call(`_iscream_Cpp_query_interval`, bedfile, regions)
}

scan_tabix <- function(bedfile, regions) {
    .Call(`_iscream_scan_tabix`, bedfile, regions)
}

#' Query all methylation info into M and coverage matrices
#' @param bedfiles A vector of BED files
#' @param regions A vector of regions
#' @param aligner The aligner used to make the WGBS BED files, only for
#' `make_mat_bsseq`
#' @param valInd The index of the data column needed for the matrix, for `make_mat`
#' @param merged Whether the input strands have been merged/collapsed
#' @param prealloc The number of rows to initialize the matrices with
#' @param nthreads Set number of threads to use overriding the
#' `"iscream.threads"` option. See `?set_threads` for more information.
#'
#' @returns A list of one or two matrices, chromosome, position, and filename vectors
#'
#' @keywords internal
Cpp_query_all <- function(bedfiles, regions, aligner, valInd, merged, sparse, prealloc, nthreads) {
    .Call(`_iscream_Cpp_query_all`, bedfiles, regions, aligner, valInd, merged, sparse, prealloc, nthreads)
}

#' Apply a function over BED file records within genomic features
#'
#' This function should be called from `summarize_regions()` since there are few
#' sanity checks on the C++ side.
#' @param bedfiles A vector of BED file paths
#' @param regions A vector of genomic regions
#' @param fun_vec Vector of the armadillo-supported stats functions to apply over the
#' CpGs in the ' regions: `"sum"`, `"mean"`, `"median"`, `"stddev"`,
#' `"variance"` "`count`", `"min"`,`"max"`, and `"range"`.
#' @param col_indices A vector of genomic regions
#' @param col_names A vector of genomic regions
#' @param mval Calculates M values when TRUE, use beta values when FALSE
#' @param region_rownames Whether to set rownames to the regions strings. Not
#' necessary if your regions vector is unnamed. If its names, then the "feature"
#' column is set to the names and the rownames are set to the regions string
#' @param nthreads Number of cores to use. See details.
#'
#' @details
#' The optimal number of threads depends on the number of bedfiles, but is set
#' to half the available OpenMP cores. See `?get_threads` for more details. It
#' can be manaully set with `set_threads()`.
#'
#' @returns A summary data.frame
#'
#' @keywords internal
Cpp_summarize_regions <- function(bedfiles, regions, fun_vec, col_indices, col_names, aligner, mval = FALSE, region_rownames = FALSE, nthreads = 1L) {
    .Call(`_iscream_Cpp_summarize_regions`, bedfiles, regions, fun_vec, col_indices, col_names, aligner, mval, region_rownames, nthreads)
}

#' Make a matrix from a numeric column of BED files
#'
#' Queries the provided regions and produces a matrix along with genomic
#' positions as a named list (`make_mat()`), a `RangedSummarizedExperiment`
#' (`make_mat_se()`), `GRanges` (`make_mat_gr()`). Parallelized across files
#' using threads from the `"iscream.threads"` option.
#'
#' @param bedfiles A vector of BED file paths
#' @param regions A vector, data frame or GenomicRanges of genomic regions. See details.
#' @param column The index of the data column needed for the matrix
#' @param mat_name What to name the matrix in the returned object
#' @param sparse Whether to return a sparse matrix
#' @param prealloc The number of rows to initialize the matrices with. If the
#' number of loci are approximately known, this can reduce runtime as fewer
#' resizes need to be made.
#' @param nthreads Set the number of threads to use. Overrides the
#' `"iscream.threads"` option. See `?set_threads` for more information.
#'
#' @returns
#' - `make_mat()`: A named list of
#'   - the matrix with the value of interest
#'   - a character vector of chromosomes and numeric vector of base positions
#'   - a character vector of the input sample BED file names
#' - `make_mat_gr()`: if `GenomicRanges` is available, a `GRanges`
#' - `make_mat_se()`: if `SummarizedExperiment` is available, a `RangedSummarizedExperiment`
#'
#' @details
#' The input regions may be string vector in the form "chr:start-end"
#' or a GRanges object. If a data frame is provided, they must have "chr",
#' "start", and "end" columns.
#'
#' @importFrom Matrix drop0
#' @importFrom methods is
#'
#' @export
#' @examples
#' bedfiles <- system.file("extdata", package = "iscream") |>
#'   list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
#' # examine the bedfiles
#' colnames <- c("chr", "start", "end", "beta", "coverage")
#' lapply(bedfiles, function(i) knitr::kable(read.table(i, col.names = colnames)))
#'
#' # make a vector of regions
#' regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
#' # make matrix of beta values
#' make_mat(bedfiles, regions, column = 4)
make_mat <- function(
  bedfiles,
  regions,
  column,
  mat_name = "value",
  sparse = FALSE,
  prealloc = 10000,
  nthreads = NULL
) {
  if (column < 4) {
    stop("`col` < 3 - must be a the index of a numeric data column not any of chr, start or end ")
  }
  verify_files_or_stop(bedfiles, verify_tabix = TRUE)
  if (is(regions, "GRanges")) {
    regions <- get_granges_string(regions)
  } else if ("data.frame" %in% class(regions)) {
    regions <- get_df_string(regions)
  }

  n_threads <- .get_threads(nthreads)
  validate_log_level(n_threads = n_threads)

  mat <- Cpp_query_all(
    bedfiles = bedfiles,
    regions = regions,
    aligner = "none",
    valInd = column,
    merged = FALSE,
    sparse = sparse,
    prealloc = prealloc,
    nthreads = n_threads
  )

  names(mat)[which(names(mat) == "M")] <- mat_name
  mat
}

#' @rdname make_mat
#' @export
make_mat_se <- function(
  bedfiles,
  regions,
  column,
  mat_name = "value",
  sparse = FALSE,
  prealloc = 10000,
  nthreads = NULL
) {
  if (!requireNamespace("SummarizedExperiment", quietly = TRUE)) {
    stop("The 'SummarizedExperiment' package must be installed for this functionality")
  }
  is_package_loaded("SummarizedExperiment", "make_mat_se", fail = TRUE)
  mat <- make_mat(bedfiles, regions, column, mat_name, sparse, prealloc, nthreads)
  matlist <- list(mat[[mat_name]])
  names(matlist) <- mat_name
  gr <- getGR(mat$chr, mat$pos)
  SummarizedExperiment::SummarizedExperiment(assays = matlist, rowRanges = gr)
}

#' @rdname make_mat
#' @export
make_mat_gr <- function(
  bedfiles,
  regions,
  column,
  mat_name = "value",
  prealloc = 10000,
  nthreads = NULL
) {
  if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
    stop("The 'GenomicRanges' package must be installed for this functionality")
  }
  is_package_loaded("GenomicRanges", "make_mat_gr", fail = TRUE)
  mat <- make_mat(
    bedfiles,
    regions,
    column,
    mat_name,
    sparse = FALSE,
    prealloc,
    nthreads
  )
  gr <- getGR(mat$chr, mat$pos)
  GenomicRanges::mcols(gr) <- mat[[mat_name]]
  gr
}

library(data.table)
library(Matrix)
extdata <- system.file("extdata", package = "iscream")
biscuit_bedfiles <- list.files(extdata, pattern = "[a|b|c|d].bed.gz$", full.names = T)
bismark_bedfiles <- list.files(extdata, pattern = "[a|b|c|d].cov.gz$", full.names = T)
regions <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14")
regions.dt <- as.data.table(regions)[, tstrsplit(regions, ":|-")]
colnames(regions.dt) <- c("chr", "start", "end")
gr <- GenomicRanges::GRanges(regions)

M <- matrix(
  c(1, 0, 0, 1, 1, 0, 2, 2, 0, 2, 0, 0, 0, 1, 0, 0, 1, 0, 1, 1, 2, 0, 0, 0, 3, 1, 0, 1),
  nrow = 7,
  byrow = TRUE
)

Cov <- matrix(
  c(1, 2, 0, 1, 1, 0, 2, 2, 2, 2, 0, 0, 1, 1, 2, 1, 2, 0, 1, 2, 2, 2, 0, 0, 3, 1, 0, 1),
  nrow = 7,
  byrow = TRUE
)

beta <- matrix(
  c(1, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0.5, 0, 1, 0.5, 1, 0, 0, 0, 1, 1, 0, 1),
  nrow = 7,
  byrow = TRUE
)

M_sp <- Matrix(M, sparse = T)
Cov_sp <- Matrix(Cov, sparse = T)
beta_sp <- Matrix(beta, sparse = T)
M_perc <- beta * 100
M_perc_sp <- Matrix(M_perc, sparse = T)

row_names <- c("chr1:2", "chr1:4", "chr1:6", "chr1:8", "chr1:10", "chr1:12", "chr1:14")
col_names <- c("a", "b", "c", "d")
colnames(M) <- col_names
colnames(Cov) <- col_names
colnames(beta) <- col_names
colnames(M_sp) <- col_names
colnames(Cov_sp) <- col_names
colnames(beta_sp) <- col_names
colnames(M_perc) <- col_names
colnames(M_perc_sp) <- col_names

exp_pos <- seq(1, 13, by = 2)
exp_seq <- rep("chr1", 7)

set_threads(1)

biscuit_bs_MC <- make_mat_bsseq(biscuit_bedfiles, regions, prealloc = 2)
bismark_bs_MC <- make_mat_bsseq(bismark_bedfiles, regions, aligner = "bismark")

biscuit_bs_MC_sp <- make_mat_bsseq(biscuit_bedfiles, regions, sparse = T)
bismark_bs_MC_sp <- make_mat_bsseq(bismark_bedfiles, regions, sparse = T, aligner = "bismark")

biscuit_bs_MC_GR <- make_mat_bsseq(biscuit_bedfiles, gr, prealloc = 2)
biscuit_bs_MC_df <- make_mat_bsseq(biscuit_bedfiles, regions.dt, prealloc = 2)

biscuit_bs_bC <- make_mat_bsseq(biscuit_bedfiles, regions, sparse = F, mval = F)
bismark_bs_bC <- make_mat_bsseq(bismark_bedfiles, regions, sparse = F, aligner = "bismark", mval = F)

biscuit_bs_bC_sp <- make_mat_bsseq(biscuit_bedfiles, regions, sparse = T, mval = F)
bismark_bs_bC_sp <- make_mat_bsseq(bismark_bedfiles, regions, sparse = T, aligner = "bismark", mval = F)

biscuit_b <- make_mat(biscuit_bedfiles, regions, column = 4, mat_name = "beta")
bismark_M <- make_mat(bismark_bedfiles, regions, column = 4, mat_name = "M")

biscuit_b_sp <- make_mat(biscuit_bedfiles, regions, column = 4, sparse = T, mat_name = "beta")
bismark_M_sp <- make_mat(bismark_bedfiles, regions, column = 4, sparse = T, mat_name = "M")

biscuit_C_gr <- make_mat(biscuit_bedfiles, gr, column = 5, prealloc = 2)
biscuit_C_df <- make_mat(biscuit_bedfiles, regions.dt, column = 5, prealloc = 2)

results <- list(
  biscuit_bs_MC,
  bismark_bs_MC,

  biscuit_bs_MC_sp,
  bismark_bs_MC_sp,

  biscuit_bs_MC_GR,
  biscuit_bs_MC_df,

  biscuit_bs_bC,
  biscuit_bs_bC,

  biscuit_bs_bC_sp,
  bismark_bs_bC_sp,

  biscuit_b,
  bismark_M,

  biscuit_b_sp,
  bismark_M_sp,

  biscuit_C_gr
)

MC_results <- list(
  biscuit_bs_MC,
  bismark_bs_MC,
  biscuit_bs_MC_sp,
  bismark_bs_MC_sp,
  biscuit_bs_MC_GR,
  biscuit_bs_MC_df
)

bC_results <- list(
  biscuit_bs_bC,
  biscuit_bs_bC,
  biscuit_bs_bC_sp,
  bismark_bs_bC_sp
)

bs_results <- c(MC_results, bC_results)

mat_results <- list(
  biscuit_b,
  bismark_M,

  biscuit_b_sp,
  bismark_M_sp,

  biscuit_C_gr
)

test_names <- function(result_obj, sparse, mat_names) {
  expected_names <- c(mat_names, c("pos", "chr", "sampleNames"))
  expect_equal(names(result_obj), expected_names)
}


bsseq_names_M <- c("M", "Cov")
bsseq_names_b <- c("beta", "Cov")

bs_MC_results <- list(
  biscuit_bs_MC,
  bismark_bs_MC,
  biscuit_bs_MC_GR,
  biscuit_bs_MC_df
)

bs_MC_sp_results <- list(
  biscuit_bs_MC_sp,
  bismark_bs_MC_sp
)

bs_bC_results <- list(
  biscuit_bs_bC,
  bismark_bs_bC
)

bs_bC_sp_results <- list(
  biscuit_bs_bC_sp,
  bismark_bs_bC_sp
)

test_that("matrix names", {
  lapply(bs_MC_results, test_names, F, bsseq_names_M)
  lapply(bs_MC_sp_results, test_names, T, bsseq_names_M)
  lapply(bs_bC_results, test_names, F, bsseq_names_b)
  lapply(bs_bC_sp_results, test_names, T, bsseq_names_b)

  test_names(biscuit_b, F, "beta")
  test_names(biscuit_b_sp, T, "beta")

  test_names(bismark_M, F, "M")
  test_names(bismark_M_sp, T, "M")
})

test_types <- function(result_obj, sparse, mat_names) {
  dense_class <- "matrix"
  sparse_class <- "dgCMatrix"

  if (sparse) {
    cl <- sparse_class
  } else {
    cl <- dense_class
  }

  for (mat in mat_names) {
    expect_equal(class(result_obj), "list")
    expect_equal(class(result_obj[[mat]])[1], cl)
  }
}

test_that("matrix metadata", {
  lapply(bs_MC_results, test_types, F, bsseq_names_M)
  lapply(bs_MC_sp_results, test_types, T, bsseq_names_M)
  lapply(bs_bC_results, test_types, F, bsseq_names_b)
  lapply(bs_bC_sp_results, test_types, T, bsseq_names_b)

  test_types(biscuit_b, F, "beta")
  test_types(biscuit_b_sp, T, "beta")

  test_types(bismark_M, F, "M")
  test_types(bismark_M_sp, T, "M")
})

test_content <- function(result_obj, sparse, mval, test_mat = NULL, mat_name = NULL) {
  if (!is.null(test_mat) & !is.null(mat_name)) {
    return(expect_equal(result_obj[[mat_name]], test_mat))
  }

  if (sparse) {
    exp_m <- if (mval) {
      M_sp
    } else {
      beta_sp
    }
    exp_c <- Cov_sp
  } else {
    exp_m <- if (mval) {
      M
    } else {
      beta
    }
    exp_c <- Cov
  }

  m_beta <- ifelse(mval, result_obj$M, result_obj$beta)
  if (mval) {
    expect_equal(result_obj$M, exp_m)
  } else {
    expect_equal(result_obj$beta, exp_m)
  }
  expect_equal(result_obj$Cov, exp_c)
}

test_that("matrix content", {
  lapply(bs_MC_results, test_content, FALSE, TRUE)
  lapply(bs_MC_results, test_content, FALSE, TRUE)
  lapply(bs_MC_sp_results, test_content, TRUE, TRUE)
  lapply(bs_bC_results, test_content, FALSE, FALSE)
  lapply(bs_bC_sp_results, test_content, TRUE, FALSE)

  test_content(biscuit_b, FALSE, test_mat = beta, mat_name = "beta")
  test_content(biscuit_b_sp, T, test_mat = beta_sp, mat_name = "beta")

  test_content(bismark_M, F, test_mat = M_perc, mat_name = "M")
  test_content(bismark_M_sp, T, test_mat = M_perc_sp, mat_name = "M")
})

exp_dims <- c(7, length(biscuit_bedfiles))
test_dims <- function(result_obj, exp_dims, mval) {
  if ("beta" %in% names(result_obj)) {
    mbeta_name <- "beta"
  } else {
    mbeta_name <- "M"
  }
  expect_equal(dim(result_obj[[mbeta_name]]), exp_dims)
  expect_equal(dim(result_obj$Cov), exp_dims)
}

test_that("matrix row,colnames,sampleNames", {
  lapply(results[1], test_dims, exp_dims)
})

suppressPackageStartupMessages({
  library(GenomicRanges)
  library(SummarizedExperiment)
})

test_that("SE, GR", {
  grmat <- make_mat_gr(biscuit_bedfiles, regions, column = 4, mat_name = "beta")
  expect_equal(class(grmat)[1], "GRanges")
  semat <- make_mat_se(biscuit_bedfiles, regions, column = 4, mat_name = "beta")
  expect_equal(class(semat)[1], "RangedSummarizedExperiment")
})

test_that("invalid columns", {
  lapply(1:3, function(i) {
    expect_error(
      make_mat(biscuit_bedfiles, regions, column = i, mat_name = "beta"),
      "`col` < 3 - must be a the index of a numeric data column not any of chr, start or end "
    )
  })
})

#' @rdname tabix
#' @param zero_based Whether the input BED file has a zero-based start column -
#' used when coverting the result data frame to GenomicRanges.
#' @export
tabix_gr <- function(
  bedfiles,
  regions,
  aligner = NULL,
  col.names = NULL,
  zero_based = TRUE,
  nthreads = NULL
) {
  if (!requireNamespace("GenomicRanges", quietly = TRUE)) {
    stop("The 'GenomicRanges' package must be installed for this functionality")
  }
  is_package_loaded("GenomicRanges", "tabix_gr", fail = TRUE)
  result <- tabix(
    bedfiles,
    regions,
    aligner,
    col.names,
    nthreads
  )
  if (is.null(result)) {
    stop("No records found")
  }
  result.gr <- GenomicRanges::makeGRangesFromDataFrame(
    result,
    starts.in.df.are.0based = zero_based,
    keep.extra.columns = TRUE
  )
  if (is(regions, "character")) {
    regions <- GenomicRanges::GRanges(regions)
  } else if ("data.frame" %in% class(regions)) {
    regions <- GenomicRanges::makeGRangesFromDataFrame(regions)
  }
  overlaps <- GenomicRanges::findOverlaps(result.gr, regions)

  if (dim(GenomicRanges::mcols(regions))[2] > 0) {
    mcols.colnames <- colnames(GenomicRanges::mcols(regions))
    mcols.subjectHits <- GenomicRanges::mcols(regions)[S4Vectors::subjectHits(overlaps), mcols.colnames]
    GenomicRanges::mcols(result.gr)[S4Vectors::queryHits(overlaps), mcols.colnames] <-
      GenomicRanges::mcols(regions)[S4Vectors::subjectHits(overlaps), mcols.colnames]
  }
  if ("file" %in% colnames(GenomicRanges::mcols(result.gr))) {
    return(GenomicRanges::split(result.gr, as.factor(result.gr$file)))
  }
  result.gr
}

## ----setup--------------------------------------------------------------------
library(iscream)
data_dir <- system.file("extdata", package = "iscream")
bedfiles <- list.files(data_dir, pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
regions <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14")

## ----gr, message=FALSE--------------------------------------------------------
if (!require("GenomicRanges", quietly = TRUE)) {
  stop("The 'GenomicRanges' package must be installed for this functionality")
}

## ----tabix_gr-----------------------------------------------------------------
gr <- GRanges(regions)
values(gr) <- DataFrame(
    gene = c("gene1", "gene2", "gene3"),
    some_metadata = c("s1", "s2", "s3")
)
gr
tabix_gr(bedfiles[1], gr)

## ----tabix--------------------------------------------------------------------
tabix(bedfiles[1], gr) |>
  makeGRangesFromDataFrame(
    starts.in.df.are.0based = TRUE,
    keep.extra.columns = TRUE
  )

## ----summarize_meth_regions---------------------------------------------------
(summary <- summarize_meth_regions(
  bedfiles,
  unname(regions),
  fun = c("sum", "mean"))
)
GRanges(summary$feature, summary = summary[, -1])

## ----summarize_regions--------------------------------------------------------
(summary <- summarize_regions(
  bedfiles,
  regions,
  column = 4,
  set_region_rownames = TRUE,
  fun = c("sum", "mean"))
)
GRanges(rownames(summary), summary = summary)

## ----make_mat-----------------------------------------------------------------
make_mat_gr(bedfiles, regions, column = 4, mat_name = "beta")

## ----mat_se, message=FALSE----------------------------------------------------
if (!require("SummarizedExperiment", quietly = TRUE)) {
  stop("The 'SummarizedExperiment' package must be installed for this functionality")
}
make_mat_se(bedfiles, regions, column = 4, mat_name = "beta", sparse = TRUE)

## ----message=FALSE------------------------------------------------------------
if (!require("bsseq", quietly = TRUE)) {
  stop("The 'bsseq' package must be installed for this functionality")
}
mats <- make_mat_bsseq(bedfiles, regions, sparse = FALSE)
do.call(BSseq, mats)

## ----si-----------------------------------------------------------------------
sessionInfo()

---
title: "iscream compatible data structures"
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{iscream compatible data structures}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
#bibliography: refs.bib
link-citations: yes
---

The examples here show how iscream output can be converted into other data
structures for further analysis.

```{r setup}
library(iscream)
data_dir <- system.file("extdata", package = "iscream")
bedfiles <- list.files(data_dir, pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
regions <- c(A = "chr1:1-6", B = "chr1:7-10", C = "chr1:11-14")
```

## `r Biocpkg("GenomicRanges")`

```{r gr, message=FALSE}
if (!require("GenomicRanges", quietly = TRUE)) {
  stop("The 'GenomicRanges' package must be installed for this functionality")
}
```

`GRanges` objects can be used as the input regions to all of iscream's functions
and can be returned by `tabix_gr()` and `make_mat_gr()`.

### From `tabix` queries

`tabix_gr()` returns a `GenomicRanges` object. The `regions` parameter can be a
string vector, a data frame or a `GRanges` object. If given a `GRanges` object
with metadata, those columns will be preserved in the output.


```{r tabix_gr}
gr <- GRanges(regions)
values(gr) <- DataFrame(
    gene = c("gene1", "gene2", "gene3"),
    some_metadata = c("s1", "s2", "s3")
)
gr
tabix_gr(bedfiles[1], gr)
```

The `data.table` output of `tabix()` can also be piped into `GRanges`, but does
not preserve input metadata.

```{r tabix}
tabix(bedfiles[1], gr) |>
  makeGRangesFromDataFrame(
    starts.in.df.are.0based = TRUE,
    keep.extra.columns = TRUE
  )
```

If the input BED file is not zero-based (e.g. Bismark coverage files), set
`zero_based = FALSE` in the `tabix()` call to get the correct conversion from
data frame to GenomicRanges.

### From `summarize_regions()`

`summarize_regions()` returns a data frame with a feature column identifying
each summary row's genomic region.

If the region features are not named (see `?summarize_regions`), pass the
`feature` column with the genomic regions as the GRanges input regions. Here,
since the `regions` vector is named, using `unname` will cause the `feature`
column to populate with regions strings:

```{r summarize_meth_regions}
(summary <- summarize_meth_regions(
  bedfiles,
  unname(regions),
  fun = c("sum", "mean"))
)
GRanges(summary$feature, summary = summary[, -1])
```

If the input regions are named use `set_region_rownames = TRUE` so that the
genomic regions strings are preserved and use them as the `GRanges` input
regions.

```{r summarize_regions}
(summary <- summarize_regions(
  bedfiles,
  regions,
  column = 4,
  set_region_rownames = TRUE,
  fun = c("sum", "mean"))
)
GRanges(rownames(summary), summary = summary)
```

### From `make_mat`

`make_mat_gr()` returns a `GRanges` object for dense matrices.

```{r make_mat}
make_mat_gr(bedfiles, regions, column = 4, mat_name = "beta")
```

## `r Biocpkg("SummarizedExperiment")`

`make_mat_se()` returns a `RangedSummarizedExperiment` for both sparse and dense
matrices.

```{r mat_se, message=FALSE}
if (!require("SummarizedExperiment", quietly = TRUE)) {
  stop("The 'SummarizedExperiment' package must be installed for this functionality")
}
make_mat_se(bedfiles, regions, column = 4, mat_name = "beta", sparse = TRUE)
```

### Making `BSseq` objects

A `r Biocpkg("bsseq")` object is a type of SummarizedExperiment, but it cannot
handle sparse matrices:

```{r, message=FALSE}
if (!require("bsseq", quietly = TRUE)) {
  stop("The 'bsseq' package must be installed for this functionality")
}
mats <- make_mat_bsseq(bedfiles, regions, sparse = FALSE)
do.call(BSseq, mats)
```

## Session info

```{r si}
sessionInfo()
```

---
title: "Plotting TSS methylation profiles"
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Plotting TSS methylation profiles}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
#bibliography: refs.bib
link-citations: yes
---



The workflow here shows how *iscream* can be used to quickly explore methylation
profiles of given genomic regions. *iscream's* tabix querying functionality can
be used to plot methylation profiles around transcription start sites (TSS).

*methscan* is a tool used to analyze single-cell bisulfite sequencing data to
find differentially methylated regions (DMRs) in the genome. The plot created
here is a reproduction of the TSS methylation profile plot made in the
[*methscan*
tutorial](https://anders-biostat.github.io/MethSCAn/tutorial.html#2-filtering-low-quality-cells)
as part of the filtering done before DMR analysis. The *methscan* workflow to
produce the plot involves three steps:

- converting the coverage BED files into Numpy sparse matrices on disk

- generating the TSS profiles from these matrices

- summarizing and plotting

Although *iscream* is not designed to run analyses on full genomes, it can be
used to explore regions such as TSS flanking regions, gene bodies, or DMRs found
using tools like *methscan* more efficiently. For example, using iscream,
all the steps above can be done in R directly from tabixed BED files for both
single-cell and bulk data. With *methscan* the first two steps are run on the
command line while the third is done in R.

## Setup


``` r
options("iscream.threads" = 8)
library(iscream)
library(data.table)
library(ggplot2)
```

### Download the data

Running this vignette requires downloading 18MB of BED files and tabix indices
from this Zenodo record: <https://zenodo.org/records/14733834>


``` r
library("BiocFileCache") |> suppressPackageStartupMessages()
cachedir <- BiocFileCache()
methscan_zip_path <- bfcrpath(cachedir, "https://zenodo.org/records/14733834/files/methscan_data.zip")
methscan_unzip <- file.path(tempdir(), "methscan")
unzip(methscan_zip_path, exdir = methscan_unzip)
methscan_dir <- file.path(methscan_unzip, "scbs_tutorial_data")
start_time = proc.time()
```

First, we generate a list of the BED file paths:


``` r
bedfiles <- list.files(
    methscan_dir,
    pattern = "*.cov.gz$",
    full.names = TRUE
)
```

## Using `tabix()`

### Get the Transcription start sites and flanking regions

Then we read the provided TSS BED file and create 2kb flanking regions around
the start sites.


``` r
tss.regions <- fread(
    file.path(methscan_dir, "Mus_musculus.GRCm38.102_TSS.bed"), drop = c(3, 5, 6)
)
colnames(tss.regions) <- c("chr", "tss", "geneID")
head(tss.regions)
```

```
##       chr     tss             geneID
##    <char>   <int>             <char>
## 1:      1 3671498 ENSMUSG00000051951
## 2:      1 4409241 ENSMUSG00000025900
## 3:      1 4496413 ENSMUSG00000025902
## 4:      1 4785739 ENSMUSG00000033845
## 5:      1 4807823 ENSMUSG00000025903
## 6:      1 4857814 ENSMUSG00000033813
```

``` r
tss.regions[, `:=`(tss.start = tss - 2000, tss.end = tss + 2000)]

# make a new data frame with chr, start, end as iscream requires these columns
tss.for_query <- tss.regions[, .(chr, start = tss.start, end = tss.end)]
```

### Make a tabix query of the TSS flanking regions

The `tabix()` function queries the provided BED files for the TSS flanking
regions to produce a data frame:


``` r
query_runtime.start <- proc.time()
tss.query <- tabix(bedfiles, tss.for_query, aligner = "bismark")
head(tss.query)
```

```
##       chr   start     end methylation.percentage count.methylated
##    <char>   <int>   <int>                  <num>            <int>
## 1:      1 4785488 4785488                      0                0
## 2:      1 4785513 4785513                      0                0
## 3:      1 4785522 4785522                      0                0
## 4:      1 4785533 4785533                      0                0
## 5:      1 4786780 4786780                    100                1
## 6:      1 4786886 4786886                    100                1
##    count.unmethylated    file
##                 <int>  <char>
## 1:                  2 cell_01
## 2:                  2 cell_01
## 3:                  2 cell_01
## 4:                  2 cell_01
## 5:                  0 cell_01
## 6:                  0 cell_01
```

### Summarize average methylation profile around TSS

Given the CpG level methylation data frame, we now
[join](https://rdatatable.gitlab.io/data.table/library/data.table/doc/datatable-joins.html)
the queried data based on CpGs that fall within the TSS flanking regions to get
the CpGs 2kb around the TSS. We can also set a new `position` column relative
to the TSS (using rounded values as in the *methscan* tutorial):


``` r
# join
tss.profile <- tss.regions[tss.query, .(
    chr,
    start,
    position = round(start - tss, -1L),
    methylation.percentage,
    file
  ),
  on = .(chr, tss.start <= start, tss.end >= end)
] |> unique()

# get mean methylation by relative position and cell
tss.summary <- tss.profile[,
  .(meth_frac = mean(methylation.percentage/100)),
  by = .(position, file)
]

query_runtime <- timetaken(query_runtime.start)
```

Time to make the query and compute the summary: 2.444s elapsed (5.262s cpu).

### Plot average methylation profiles around the TSS


``` r
tss.plot <- ggplot(tss.summary, aes(x = position / 1000, y = meth_frac)) +
  scale_y_continuous(
    labels=scales::percent_format(accuracy=1),
    limits=c(0, 1), breaks=c(0, .5, 1)
  ) +
  geom_line(linewidth = .1) +
  facet_wrap(~file) +
  labs(x = "position relative to TSS [kb]", y = "DNA methylation")

total_runtime <- timetaken(start_time)

tss.plot
```

<div class="figure" style="text-align: center">
<img src="../man/figures/tss_plot-1.png" alt="TSS profiles"  />
<p class="caption">TSS profiles</p>
</div>


Total runtime, from getting the BED files and regions to making the query,
calculating the summaries and plotting: 2.489s elapsed (5.307s cpu). With methscan,
generating the TSS methylation profiles alone took 11 seconds.

## Using `summarize_regions`

A similar analysis can be done using the `summarize_regions()` function if you
only need to see the distribution of beta means by file, rather than means by
relative position per file


``` r
library("ggridges")
tss.means <- summarize_meth_regions(
  bedfiles,
  tss.for_query,
  aligner = "bismark",
  fun = "mean",
  mval = FALSE
) |> as.data.table()
```

```
## [11:28:24.770899] [iscream::summarize_regions] [info] Summarizing 21622 regions from 30 bedfiles
## [11:28:24.770922] [iscream::summarize_regions] [info] using mean
## [11:28:24.770925] [iscream::summarize_regions] [info] with columns 4, 5 as coverage, beta
```

``` r
ggplot(tss.means, aes(x = beta.mean, y = file, fill = after_stat(x))) +
  geom_density_ridges_gradient() +
  scale_fill_distiller(palette = "BrBG") +
  theme_classic()
```

<div class="figure" style="text-align: center">
<img src="../man/figures/summarize_regions-1.png" alt="TSS distribution by file"  />
<p class="caption">TSS distribution by file</p>
</div>

For per-file means you could collapse the means within file:


``` r
tss.means[, .(beta.mean = mean(beta.mean, na.rm = TRUE)), by = file] |>
  ggplot(
    aes(
      x = reorder(file, beta.mean),
      y = beta.mean)
    ) +
  geom_bar(stat = 'identity') +
  theme_classic() +
  coord_flip() +
  labs(x = "Sample")
```

<div class="figure" style="text-align: center">
<img src="../man/figures/unnamed-chunk-6-1.png" alt="Mean TSS by file"  />
<p class="caption">Mean TSS by file</p>
</div>


## Session info


``` r
sessionInfo()
```

```
## R version 4.5.1 (2025-06-13)
## Platform: x86_64-pc-linux-gnu
## Running under: Ubuntu 24.04.2 LTS
## 
## Matrix products: default
## BLAS/LAPACK: /nix/store/yf6dpab0gcjr9gvpww1zlafs9n0f48h3-blas-3/lib/libblas.so.3;  LAPACK version 3.12.0
## 
## locale:
##  [1] LC_CTYPE=en_US.UTF-8       LC_NUMERIC=C              
##  [3] LC_TIME=en_US.UTF-8        LC_COLLATE=en_US.UTF-8    
##  [5] LC_MONETARY=en_US.UTF-8    LC_MESSAGES=en_US.UTF-8   
##  [7] LC_PAPER=en_US.UTF-8       LC_NAME=C                 
##  [9] LC_ADDRESS=C               LC_TELEPHONE=C            
## [11] LC_MEASUREMENT=en_US.UTF-8 LC_IDENTIFICATION=C       
## 
## time zone: America/Detroit
## tzcode source: system (glibc)
## 
## attached base packages:
## [1] stats     graphics  grDevices utils     datasets  methods   base     
## 
## other attached packages:
## [1] ggridges_0.5.7       BiocFileCache_2.99.6 dbplyr_2.5.1        
## [4] ggplot2_4.0.0        data.table_1.17.8    iscream_0.99.9      
## 
## loaded via a namespace (and not attached):
##  [1] Matrix_1.7-4          bit_4.6.0             gtable_0.3.6         
##  [4] dplyr_1.1.4           compiler_4.5.1        filelock_1.0.3       
##  [7] tidyselect_1.2.1      Rcpp_1.1.0            blob_1.2.4           
## [10] parallel_4.5.1        scales_1.4.0          fastmap_1.2.0        
## [13] lattice_0.22-7        R6_2.6.1              labeling_0.4.3       
## [16] generics_0.1.4        curl_7.0.0            httr2_1.2.1          
## [19] knitr_1.50            tibble_3.3.0          stringfish_0.17.0    
## [22] DBI_1.2.3             pillar_1.11.1         RColorBrewer_1.1-3   
## [25] rlang_1.1.6           cachem_1.1.0          xfun_0.53            
## [28] S7_0.2.0              bit64_4.6.0-1         RcppParallel_5.1.11-1
## [31] memoise_2.0.1         RSQLite_2.4.3         cli_3.6.5            
## [34] withr_3.0.2           magrittr_2.0.4        grid_4.5.1           
## [37] pbapply_1.7-4         rappdirs_0.3.3        lifecycle_1.0.4      
## [40] vctrs_0.6.5           evaluate_1.0.5        glue_1.8.0           
## [43] farver_2.1.2          parallelly_1.45.1     purrr_1.1.0          
## [46] tools_4.5.1           pkgconfig_2.0.3
```

<!-- vim: set filetype=rmd: -->

---
title: "An introduction to iscream"
author:
- name: James Eapen
  affiliation:
  - Department of Epigenetics, Van Andel Institute, Grand Rapids, MI
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{An introduction to iscream}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: refs.bib
link-citations: yes
---

```{r setup, include=FALSE}
options("datatable.print.nrows" = 50)
```

## Introduction

iscream is a BED file querying package that can can retrieve records within
regions of interest from multiple [tabixed](https://en.wikipedia.org/wiki/Tabix)
BED files. It can make queries like tabix, summarize data within regions and
make matrices across input files and regions. All operations can be done in
parallel and return R/Bioconductor data structures.

iscream was designed to be memory efficient and fast on large datasets. This
vignette demonstrates iscream's functions on a small dataset. For more
information on performance and considerations on larger real datasets see
`vignette("performance")`.

## Installation

iscream may be installed from <https://bioconductor.org> with

```{r install, eval=FALSE}
if (!require("BiocManager"))
    install.packages("BiocManager")
BiocManager::install("iscream")
```

Although iscream will compile using `Rhtslib`, it will perform best if compiled
against [htslib](https://www.htslib.org/download/) which has itself been
compiled with `libdeflate`. See `vignette("htslib")` for more information.

### Loading iscream

```{r load_iscream}
library(iscream)
set_threads(2)
```

On load, iscream will inform the user about the number if threads it is set to
use. This is configurable with either `set_threads()` or by setting
`option("iscream.threads")`, which can be done in your `.Rprofile` to set it on
startup. Once loaded you use the following querying functions.

## `tabix()`

tabix is a command-line tool that queries records from compressed BED files. As
input it receives the path to a tabixed BED file and one or more genomic regions
of interest. It queries and prints the records from the tabixed BED file that
fall within the regions of interest. This output may be redirected to a file and
read into R.

iscream's `tabix()` function works in a similar way but supports multiple files
and multiple input formats all in one function. It receives a vector of paths to
tabixed BED files and regions of interest. The input regions may be a vector of
`"chr:start-end"` strings, a data frame with `"chr"`, `"start"`, and `"end"`
columns or a `GRanges` object. It queries the BED files and returns the records
as a parsed `r CRANpkg("data.table")`. To get a `GRanges` object instead, use
`tabix_gr`.

Using iscream eliminates context switching between the shell and R for tabix
queries since the queries are made in R and the result is an R data structure.
`r Biocpkg("Rsamtools")` is another R package that can query data from BED
files, but it only supports one file at a time, only accepts `GRanges` as the
input regions and does not parse the tab-delimited strings. See
`vignette("tabix")` for a comparison of iscream and Rsamtools.

### Setup

Using `list.files()`, you can get the paths to all compressed bed files in a
directory. These files contain small regions from chromosome 1 and Y from the
snmC-seq2 methylation data [@luo2018a].

```{r load_data}
data_dir <- system.file("extdata", package = "iscream")
(bedfiles <- list.files(
  data_dir,
  pattern = "cell[1-5].bed.gz$",
  full.names = TRUE))
```

For demonstration, I'm using two regions.

```{r make_regions}
regions <- c("chr1:184577-680065", "chrY:56877780-56882524")
```

**Note**: in this dataset the chromosome identifier is in a "chr[number]"
format. However, in some datasets, the chromosome ID is just a number or a
letter like '1', or 'Y'. To see the chromosome ID format across your files you
can use `query_chroms()` to get all unique chromosomes across all input files:

```{r query_chroms}
query_chroms(bedfiles)
```

### Making queries

`tabix()` one file:

```{r tabix_one}
tabix(bedfiles[1], regions)
```

With multiple files, the output contains a column for the file name.

```{r tabix_all}
tabix(bedfiles, regions)
```

### Using `r Biocpkg("GenomicRanges")`

Both `tabix()` and `tabix_gr()` accept `GRanges` objects as input regions but
`tabix_gr()` returns `GRanges` objects instead of data frames. `tabix_gr` will
also preserve any input `GRanges` metadata.

```{r tabix_gr, message=FALSE}
if (!require("GenomicRanges", quietly = TRUE)) {
  stop("The 'GenomicRanges' package must be installed for this functionality")
}
gr <- GRanges(regions)
tabix_gr(bedfiles, gr)
```

### Setting column names

You can set the result data frame's column names or the `GRanges` `mcols` with
`col.names`:

```{r tabix_colnames}
tabix_gr(bedfiles, regions, col.names = c("beta", "coverage"))
```

### Rsamtools-style output

`tabix_raw()` will return an unparsed named list like `r Biocpkg("Rsamtools")`,
but with multi-file support:

```{r tabix_raw}
tabix_raw(bedfiles, regions)
```

### WGBS BED files

Since iscream was originally developed to read Whole Genome Bisulfite Sequencing
data, it has the `aligner` argument for automatically setting column names for
the BISCUIT [@zhou2024], Bismark [@krueger2011], and BSBolt [@farrell2021].

```{r tabix_wgbs}
tabix_gr(bedfiles, regions, aligner = "biscuit")
```

## `summarize_regions()`

Using the same BED files and regions, iscream can also summarize the data within
regions. This requires providing the index of the data columns you want
summarized and the column names for the output. All summarizing functions are
run on each input column by default - see `?summarize_regions` for supported
functions.

```{r summary}
summarize_regions(
  bedfiles,
  regions,
  columns = c(4, 5),
  col_names = c("beta", "coverage")
)
```

The `feature` column here contains the genomic region coordinates, but can be
set to something more informational if you have names for the regions. You can
also select the functions applied with `fun`:


```{r summary_named}
names(regions) <- c("R1", "R2")
summarize_regions(
  bedfiles,
  regions,
  fun = c("mean", "sum"),
  columns = 5,
  col_names = "coverage"
)
```

### WGBS BED files

For WGBS data specifically, you can use `summarize_meth_regions()`, which takes
the `aligner` argument to correctly parse the columns.

```{r summarize_meth_regions}
summarize_meth_regions(
  bedfiles,
  regions,
  aligner = "biscuit",
  fun = c("mean", "sum")
)
```


## `make_mat()`

`make_mat()` makes a matrix where every row is a locus found in at least one input
file and the columns are the input files. It returns a named list of the matrix,
coordinates, and file names you can use for analysis or other data structures.
`make_mat_se()` returns a `RangedSummarizedExperiment`, and `make_mat_gr()`
returns a `GRanges` object.

```{r make_mat, message=FALSE}
if (!require("SummarizedExperiment", quietly = TRUE)) {
  stop("The 'SummarizedExperiment' package must be installed for this functionality")
}
(mat <- make_mat_se(bedfiles, regions, column = 4, mat_name = "beta"))
head(assay(mat), 10)
```

If you have sparse data, you can save memory with `sparse = TRUE`, but only for
`make_mat()` and `make_mat_se()`.

```{r sparse}
mat <- make_mat(bedfiles, regions, column = 4, mat_name = "beta", sparse = TRUE)
head(mat$beta, 10)
```

### WGBS BED files

For a `r Biocpkg("BSseq")` compatible object, use `make_meth_mat` which returns
a list of BSseq inputs. To make the `BSseq` object run

```{r, message=FALSE}
if (require("bsseq", quietly = TRUE)) {
  meth_mat <- make_mat_bsseq(bedfiles, regions)
  do.call(BSseq, meth_mat)
}
```

## Further reading

While this vignette used a toy dataset, `vignette("performance")` has tips on
getting the best performance from iscream using a real and larger dataset with
100 files. `vignette("TSS")` is an example of iscream's application to examine
methylation around transcription start site on another real dataset. For more
information on the supported Bioconductor data structures and conversions see
`vignette("data_structures")`.

## Session info

```{r si}
sessionInfo()
```

## References

---
title: "Manuscript data availabiliy"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Manuscript data availabiliy}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: refs.bib
link-citations: yes
---

## Datasets

Two datasets were used to produce the figure in the paper:

1. snmC-seq2 scWGBS data [@luo2018a] for single-cell benchmarks.
2. bulk WGBS from a project in our lab

Both datasets were aligned with BISCUIT [@zhou2024]. To produce Bismark BEDgraph
files [@krueger2011], we used a python script to convert the beta and coverage
value columns to percentage methylation, unmethylated reads and methylated read
columns.

```py
import sys
import gzip

bedfile = sys.argv[1]

def bed2cov(line):
    chr, start, end, beta, cov = line.split('\t')
    percent, meth, unmeth = convert(float(beta), int(cov))
    return '\t'.join([chr, start, start, str(percent), str(meth), str(unmeth)])

def convert(beta, cov):
    percent = round(beta * 100)
    meth = round(beta * cov)
    unmeth = cov - meth
    return [percent, meth, unmeth]

with gzip.open(bedfile, 'rt') as bed:
    for line in bed:
        print(bed2cov(line))
```

Using GNU parallel:

```sh
parallel python bed2cov.py <biscuit_path>/{} '>' <bismark_path>/{/.} ::: biscuit_path/*.bed.gz
```

Then using the `iscream.paper` package at
<https://github.com/huishenlab/iscream.paper> we ran the benchmarks and produced
the figures.

## References

---
title: "iscream vs Rsamtools::scanTabix"
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{iscream vs Rsamtools::scanTabix}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
link-citations: yes
---



Both iscream and Rsamtools can be used to query records from tabixed BED files
in R. Here we compare their usability and performance using the methscan
dataset.

This vignette uses mouse WGBS data from *methscan* as in the [Plotting TSS
profiles tutorial](./TSS.Rmd) and mouse promoter regions. Running it requires
downloading 18MB of BED files and tabix indices from this Zenodo record:
<https://zenodo.org/records/14733834>


``` r
library("BiocFileCache") |> suppressPackageStartupMessages()
cachedir <- BiocFileCache()
methscan_zip_path <- bfcrpath(cachedir, "https://zenodo.org/records/14733834/files/methscan_data.zip")
methscan_unzip <- file.path(tempdir(), "methscan")
unzip(methscan_zip_path, exdir = methscan_unzip)
methscan_dir <- file.path(methscan_unzip, "scbs_tutorial_data")
```

iscream normally uses htslib to query BED files and store the data in memory.
However, iscream's `tabix()` function can use the tabix executable to make tabix
queries because the shell program's stream-to-file approach is faster than
allocating and storing strings in memory. By default, the `"tabix.method"`
option is set to "shell" and iscream's `tabix()` will look for the tabix
executable, falling back to using htslib only if it's not found. Here, the
`"tabix.method"` option is set to "htslib" to make fair comparisons between
Rsamtools and iscream since `scanTabix` does not use the executable, but stores
the strings in memory.


``` r
library(iscream)
```

```
## iscream using 1 thread by default but parallelly::availableCores() detects 16 possibly available threads. See `?set_threads` for information on multithreading before trying to use more.
```

``` r
options("tabix.method" = 'htslib')
options("iscream.threads" = 8)
```

Using `scanTabix` requires the input regions to be `GRanges`. `iscream::tabix`
accepts strings, data frames and GRanges.


``` r
library(data.table)
library(GenomicRanges) |> suppressPackageStartupMessages()
library(Rsamtools) |> suppressPackageStartupMessages()
library(microbenchmark)
library(pbapply)
library(ggplot2)
```


``` r
methscan_files <- list.files(
  methscan_dir,
  full.names = T,
  pattern = "*.cov.gz$"
)
mouse_promoters <- fread(file.path(methscan_dir, "mouse_promoters.bed"), col.names = c("chr", "start", "end"))
mouse_promoters.gr <- GRanges(mouse_promoters)
```


# One file

`tabix_raw()` and `scanTabix` produce a list of unparsed or raw strings. The
result of both are identical.


``` r
bench_1_raw <- microbenchmark(
  `rsamtools 1 file raw` = rq <- scanTabix(methscan_files[1], param = mouse_promoters.gr),
  `iscream 1 file raw` = iq <- tabix_raw(methscan_files[1], mouse_promoters),
  times = 3
)
bench_1_raw
```

```
## Unit: milliseconds
##                  expr        min         lq      mean     median         uq
##  rsamtools 1 file raw 2760.55888 2765.35353 2769.3114 2770.14819 2773.68760
##    iscream 1 file raw   79.47893   80.34265   84.7329   81.20637   87.35989
##         max neval
##  2777.22701     3
##    93.51341     3
```

``` r
autoplot(bench_1_raw)
```

```
## Warning: `aes_string()` was deprecated in ggplot2 3.0.0.
## ℹ Please use tidy evaluation idioms with `aes()`.
## ℹ See also `vignette("ggplot2-in-packages")` for more information.
## ℹ The deprecated feature was likely used in the microbenchmark package.
##   Please report the issue at <https://github.com/joshuaulrich/microbenchmark/issues/>.
## This warning is displayed once every 8 hours.
## Call `lifecycle::last_lifecycle_warnings()` to see where this warning was generated.
```

<div class="figure" style="text-align: center">
<img src="../man/figures/bench_1_raw-1.png" alt="tabix vs scanTabix raw string output on 1 file"  />
<p class="caption">tabix vs scanTabix raw string output on 1 file</p>
</div>

``` r
iq[1:5]
```

```
## $`1:3669498-3673498`
## character(0)
## 
## $`1:4407241-4411241`
## character(0)
## 
## $`1:4494413-4498413`
## character(0)
## 
## $`1:4783739-4787739`
## [1] "1\t4785488\t4785488\t0.000000\t0\t2"  
## [2] "1\t4785513\t4785513\t0.000000\t0\t2"  
## [3] "1\t4785522\t4785522\t0.000000\t0\t2"  
## [4] "1\t4785533\t4785533\t0.000000\t0\t2"  
## [5] "1\t4786780\t4786780\t100.000000\t1\t0"
## [6] "1\t4786886\t4786886\t100.000000\t1\t0"
## [7] "1\t4786958\t4786958\t100.000000\t1\t0"
## 
## $`1:4805823-4809823`
## [1] "1\t4806176\t4806176\t100.000000\t1\t0"
## [2] "1\t4806221\t4806221\t100.000000\t1\t0"
## [3] "1\t4807572\t4807572\t0.000000\t0\t1"  
## [4] "1\t4807669\t4807669\t0.000000\t0\t2"  
## [5] "1\t4807682\t4807682\t0.000000\t0\t2"  
## [6] "1\t4807872\t4807872\t0.000000\t0\t1"  
## [7] "1\t4807890\t4807890\t0.000000\t0\t1"  
## [8] "1\t4807940\t4807940\t0.000000\t0\t1"  
## [9] "1\t4807950\t4807950\t0.000000\t0\t1"
```

``` r
all.equal(iq, rq)
```

```
## [1] TRUE
```


# Multiple files

`tabix()` can also take multiple files in the same call. It returns the raw
strings, but each input file is a list of it's own. `scanTabix()` does not
support this.


``` r
bench_30_raw <- microbenchmark(
  `iscream 30 files raw` = iq <- tabix_raw(methscan_files, mouse_promoters),
  times = 3
)
bench_30_raw
```

```
## Unit: seconds
##                  expr      min       lq     mean   median       uq      max
##  iscream 30 files raw 1.346655 1.559583 1.859358 1.772512 2.115709 2.458907
##  neval
##      3
```

``` r
names(iq)
```

```
##  [1] "cell_01" "cell_02" "cell_03" "cell_04" "cell_05" "cell_06" "cell_07"
##  [8] "cell_08" "cell_09" "cell_10" "cell_11" "cell_12" "cell_13" "cell_14"
## [15] "cell_15" "cell_16" "cell_17" "cell_18" "cell_19" "cell_20" "cell_21"
## [22] "cell_22" "cell_23" "cell_24" "cell_25" "cell_26" "cell_27" "cell_28"
## [29] "cell_29" "cell_30"
```

``` r
iq[["cell_01"]][1:5]
```

```
## $`1:3669498-3673498`
## character(0)
## 
## $`1:4407241-4411241`
## character(0)
## 
## $`1:4494413-4498413`
## character(0)
## 
## $`1:4783739-4787739`
## [1] "1\t4785488\t4785488\t0.000000\t0\t2"  
## [2] "1\t4785513\t4785513\t0.000000\t0\t2"  
## [3] "1\t4785522\t4785522\t0.000000\t0\t2"  
## [4] "1\t4785533\t4785533\t0.000000\t0\t2"  
## [5] "1\t4786780\t4786780\t100.000000\t1\t0"
## [6] "1\t4786886\t4786886\t100.000000\t1\t0"
## [7] "1\t4786958\t4786958\t100.000000\t1\t0"
## 
## $`1:4805823-4809823`
## [1] "1\t4806176\t4806176\t100.000000\t1\t0"
## [2] "1\t4806221\t4806221\t100.000000\t1\t0"
## [3] "1\t4807572\t4807572\t0.000000\t0\t1"  
## [4] "1\t4807669\t4807669\t0.000000\t0\t2"  
## [5] "1\t4807682\t4807682\t0.000000\t0\t2"  
## [6] "1\t4807872\t4807872\t0.000000\t0\t1"  
## [7] "1\t4807890\t4807890\t0.000000\t0\t1"  
## [8] "1\t4807940\t4807940\t0.000000\t0\t1"  
## [9] "1\t4807950\t4807950\t0.000000\t0\t1"
```

``` r
scanTabix(methscan_files, param = GRanges(mouse_promoters))
```

```
## Error in h(simpleError(msg, call)): error in evaluating the argument 'file' in selecting a method for function 'scanTabix': 'file' must be length 1
```

# Parsing records into a data frame

A parsed data frame is more useful than a list of raw strings - use `tabix()`
instead of `tabix_raw()`. For `scanTabix`, this custom function parses the list
of strings to make a similar data frame:


``` r
rtbx <- function(bed) {
  q <- scanTabix(bed, param = GRanges(mouse_promoters)) |>
    lapply(strsplit, split = "\t") |>
    Filter(function(i) length(i) != 0, x = _) |>
    unlist(recursive = FALSE) |>
    do.call(rbind, args = _) |>
    as.data.table() |>
    setnames(c("chr", "start", "end", paste0("V", 1:3)))
  q
}

bench_1_df <- microbenchmark(
  `rsamtools 1 file data frame` = rq <- rtbx(methscan_files[1]),
  `iscream 1 file data frame` = iq <- tabix(methscan_files[1], mouse_promoters),
  times = 3
)
rq
```

```
##           chr     start       end         V1     V2     V3
##        <char>    <char>    <char>     <char> <char> <char>
##     1:      1   4785488   4785488   0.000000      0      2
##     2:      1   4785513   4785513   0.000000      0      2
##     3:      1   4785522   4785522   0.000000      0      2
##     4:      1   4785533   4785533   0.000000      0      2
##     5:      1   4786780   4786780 100.000000      1      0
##    ---                                                    
## 79900:      X 168673566 168673566   0.000000      0      1
## 79901:      X 168673577 168673577   0.000000      0      1
## 79902:      X 168674670 168674670   0.000000      0      1
## 79903:      X 168675047 168675047   0.000000      0      1
## 79904:      X 169318831 169318831 100.000000      1      0
```

``` r
iq
```

```
##           chr     start       end    V1    V2    V3
##        <char>     <int>     <int> <num> <int> <int>
##     1:      1   4785488   4785488     0     0     2
##     2:      1   4785513   4785513     0     0     2
##     3:      1   4785522   4785522     0     0     2
##     4:      1   4785533   4785533     0     0     2
##     5:      1   4786780   4786780   100     1     0
##    ---                                             
## 79900:      X 168673566 168673566     0     0     1
## 79901:      X 168673577 168673577     0     0     1
## 79902:      X 168674670 168674670     0     0     1
## 79903:      X 168675047 168675047     0     0     1
## 79904:      X 169318831 169318831   100     1     0
```

``` r
all.equal(iq, rq, check.attributes = F)
```

```
## [1] "Datasets have different column modes. First 3: start(numeric!=character) end(numeric!=character) V1(numeric!=character)"
```

The column types are different but the data is identical.


``` r
bench_1_df
```

```
## Unit: milliseconds
##                         expr       min       lq      mean    median        uq
##  rsamtools 1 file data frame 2169.4535 2229.954 2641.8171 2290.4544 2877.9989
##    iscream 1 file data frame  132.9694  134.259  141.7589  135.5487  146.1537
##        max neval
##  3465.5434     3
##   156.7588     3
```

``` r
autoplot(bench_1_df)
```

<div class="figure" style="text-align: center">
<img src="../man/figures/bench_1_df-1.png" alt="tabix vs scanTabix parsed data frame from 1 file"  />
<p class="caption">tabix vs scanTabix parsed data frame from 1 file</p>
</div>

# Multiple files as data frame

We can try to query multiple BED files using Rsamtools with this function that
8 cores like iscream does:


``` r
partbx <- function(bedfiles) {
  pblapply(
    methscan_files,
    function(bed) {
      query <- rtbx(bed)[, file := tools::file_path_sans_ext(basename(bed), compression = TRUE)][]
      setnames(query, c("chr", "start", "end", "beta", "M", "U", "file"))
      query
    },
    cl = 8
  ) |>
    rbindlist()
}

bench_30_df <- microbenchmark(
  `rsamtools 30 file data frame` = rq <- partbx(methscan_files),
  `iscream 30 file data frame` = iq <- tabix(methscan_files, mouse_promoters, col.names = c("beta", "M", "U")),
  times = 3
)
bench_30_df
```

```
## Unit: seconds
##                          expr       min        lq      mean   median        uq
##  rsamtools 30 file data frame 24.823254 26.949417 28.924390 29.07558 30.974959
##    iscream 30 file data frame  1.074239  1.389884  1.660032  1.70553  1.952928
##        max neval
##  32.874338     3
##   2.200326     3
```

``` r
autoplot(bench_30_df)
```

<div class="figure" style="text-align: center">
<img src="../man/figures/bench_30_df-1.png" alt="tabix vs scanTabix parsed data frame from 30 files"  />
<p class="caption">tabix vs scanTabix parsed data frame from 30 files</p>
</div>


## All benchmarks {.tabset}


``` r
bench_all <- rbind(bench_1_raw, bench_30_raw, bench_1_df, bench_30_raw, bench_30_df)

bench.dt <- as.data.table(bench_all)[, .(
  time       = time / 1000000,
  package    = fifelse(grepl("iscream", expr), "iscream", "rsamtools"),
  files      = fifelse(grepl("1", expr), 1, 30),
  data.frame = fifelse(grepl("raw", expr), FALSE, TRUE)
)][,
  `data structure` := fifelse(data.frame, "data.frame", "raw-strings")
]
```


### Runtime


``` r
ggplot(bench.dt, aes(x = as.factor(files), y = time / 1000, color = as.factor(package))) +
  geom_boxplot(position = position_dodge(preserve = "single")) +
  scale_color_discrete(name = "Package") +
  scale_y_continuous(limits = c(0, 35), breaks = scales::extended_breaks(10)) +
  facet_wrap(~ `data structure`, labeller = "label_both") +
  labs(x = "File count", y = "Runtime (seconds)") +
  theme_bw()
```

```
## Warning: Removed 1 row containing non-finite outside the scale range
## (`stat_boxplot()`).
```

<div class="figure" style="text-align: center">
<img src="../man/figures/all-1.png" alt="Comparing all benchmarked iscream and Rsamtools querying runtimes"  />
<p class="caption">Comparing all benchmarked iscream and Rsamtools querying runtimes</p>
</div>

## Session info


``` r
sessionInfo()
```

```
## R version 4.5.1 (2025-06-13)
## Platform: x86_64-pc-linux-gnu
## Running under: Ubuntu 24.04.3 LTS
## 
## Matrix products: default
## BLAS/LAPACK: /nix/store/g5b7qnbhjmiac5q4jb5yvahy6pysd88g-blas-3/lib/libblas.so.3;  LAPACK version 3.12.0
## 
## locale:
##  [1] LC_CTYPE=en_US.UTF-8       LC_NUMERIC=C              
##  [3] LC_TIME=en_US.UTF-8        LC_COLLATE=en_US.UTF-8    
##  [5] LC_MONETARY=en_US.UTF-8    LC_MESSAGES=en_US.UTF-8   
##  [7] LC_PAPER=en_US.UTF-8       LC_NAME=C                 
##  [9] LC_ADDRESS=C               LC_TELEPHONE=C            
## [11] LC_MEASUREMENT=en_US.UTF-8 LC_IDENTIFICATION=C       
## 
## time zone: America/Detroit
## tzcode source: system (glibc)
## 
## attached base packages:
## [1] stats4    stats     graphics  grDevices utils     datasets  methods  
## [8] base     
## 
## other attached packages:
##  [1] ggplot2_4.0.0        pbapply_1.7-4        microbenchmark_1.5.0
##  [4] Rsamtools_2.25.3     Biostrings_2.77.2    XVector_0.49.1      
##  [7] GenomicRanges_1.61.5 Seqinfo_0.99.2       IRanges_2.43.5      
## [10] S4Vectors_0.47.4     BiocGenerics_0.55.1  generics_0.1.4      
## [13] data.table_1.17.8    iscream_0.99.9       BiocFileCache_2.99.6
## [16] dbplyr_2.5.1        
## 
## loaded via a namespace (and not attached):
##  [1] rappdirs_0.3.3        bitops_1.0-9          RSQLite_2.4.3        
##  [4] lattice_0.22-7        magrittr_2.0.4        RColorBrewer_1.1-3   
##  [7] evaluate_1.0.5        grid_4.5.1            fastmap_1.2.0        
## [10] blob_1.2.4            Matrix_1.7-4          DBI_1.2.3            
## [13] purrr_1.1.0           scales_1.4.0          codetools_0.2-20     
## [16] httr2_1.2.1           cli_3.6.5             rlang_1.1.6          
## [19] crayon_1.5.3          parallelly_1.45.1     bit64_4.6.0-1        
## [22] withr_3.0.2           cachem_1.1.0          tools_4.5.1          
## [25] parallel_4.5.1        BiocParallel_1.43.4   memoise_2.0.1        
## [28] dplyr_1.1.4           filelock_1.0.3        curl_7.0.0           
## [31] vctrs_0.6.5           R6_2.6.1              lifecycle_1.0.4      
## [34] stringfish_0.17.0     bit_4.6.0             pkgconfig_2.0.3      
## [37] gtable_0.3.6          RcppParallel_5.1.11-1 pillar_1.11.1        
## [40] glue_1.8.0            Rcpp_1.1.0            xfun_0.53            
## [43] tibble_3.3.0          tidyselect_1.2.1      knitr_1.50           
## [46] farver_2.1.2          labeling_0.4.3        compiler_4.5.1       
## [49] S7_0.2.0
```

<!-- vim: set filetype=rmd: -->

---
title: "Getting htslib headers"
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Getting htslib headers}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

iscream currently has Rhtslib as a "LinkingTo" dependency so if a system
installation of htslib is not found with pkg-config, the installer will fall
back to using Rhtslib as the htslib header source. However, we recommend getting
a more up-to-date htslib from another source. You or your system administrator
can install htslib with the system package manager which usually sets
`PKG_CONFIG_PATH` automatically. On MacOS you can get htslib with the Homebrew
package manager. On HPC systems, htslib may be provided as a module. Make sure
these methods also set the `PKG_CONFIG_PATH`.

If you aren't able to install htslib development libraries system-wide for lack
of admin permissions, you can install them from other channels. We recommend
[pixi](#pixi) or [conda](#conda) since their htslib is compiled with
[*libdeflate*](https://github.com/ebiggers/libdeflate) support and is faster
than htslib without *libdeflate*. If you're compiling your own htslib, compile
*libdeflate* first and then htslib. With Rhtslib we've seen poorer performance
compared to a standard htslib installation, both with and without *libdeflate*.

To see what htslib version iscream is using and whether it has libdeflate, run

```r
library(iscream)
htslib_version()
#> 1.21
#> build=configure libcurl=yes S3=yes GCS=yes libdeflate=yes lzma=yes bzip2=yes plugins=no htscodecs=1.6.1
```

and check that `libdeflate=yes`.


```{r, echo=FALSE, error=FALSE}
if (!require("ggplot2", quietly = TRUE)) {
  stop("The 'ggplot2' package must be installed for this functionality")
}
```

```{r, echo=FALSE, dpi=300, fig.width=6, fig.height=4, fig.align="center", fig.cap="Effect of htslib v1.18 source on iscream's 'make_mat()' runtime from one bulk WGBS BED file"}
data_dir <- system.file("extdata", package = "iscream")
df <- read.csv(file.path(data_dir, "htslib.csv"))
ggplot(df, aes(x = regions, y = time, color = source)) +
  geom_point(alpha = 0.8) +
  stat_summary(
    aes(x = regions, group = source),
    fun = "mean",
    geom = "line",
    linewidth = 0.5,
    show.legend = TRUE
  ) +
  labs(
    x = "Genomic regions",
    y = "Runtime (s)",
    color = "htslib source"
  ) +
  theme_bw()


```

### Conda/miniconda/mamba/micromamba

Create an `environment.yaml` with the following contents to install htslib 1.21:

```yaml
name: iscream
channels:
  - bioconda
  - conda-forge
dependencies:
  - htslib=1.21
  - pkg-config=0.29.2
```

Add this file to your project directory and run

```bash
conda env create -f environment.yaml
conda activate iscream
export PKG_CONFIG_PATH=$CONDA_PREFIX/lib
export LD_LIBRARY_PATH=$CONDA_PREFIX/lib/
```

Confirm that the headers are available for compilation
```bash
pkg-config --cflags --libs htslib
```

You should get something like

```
-I/home/user/miniconda3/envs/iscream/include -L/home/user/miniconda3/envs/iscream/lib -lhts
```

### [Pixi](https://pixi.sh/latest/)

Pixi uses the conda repositories to install packages. Create a `pixi.toml` file
with this content and add the file to your project directory:

```toml
[project]
channels = ["conda-forge", "bioconda"]
name = "iscream"
platforms = ["linux-64"]
version = "0.1.0"

[activation.env]
LD_LIBRARY_PATH="$CONDA_PREFIX/lib"

[dependencies]
htslib = "1.21.*"
pkg-config = ">=0.29.2,<0.30"
```

To create an environment with the required system dependencies run

```bash
pixi shell
```

Confirm that the headers are available for compilation
```bash
pkg-config --cflags --libs htslib
```

You should get something like

```
-I/home/user/iscream/.pixi/envs/default/include -L/home/user/iscream/.pixi/envs/default/lib -lhts
```

## Session info

```{r si}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_mat.R
\name{make_mat}
\alias{make_mat}
\alias{make_mat_se}
\alias{make_mat_gr}
\title{Make a matrix from a numeric column of BED files}
\usage{
make_mat(
  bedfiles,
  regions,
  column,
  mat_name = "value",
  sparse = FALSE,
  prealloc = 10000,
  nthreads = NULL
)

make_mat_se(
  bedfiles,
  regions,
  column,
  mat_name = "value",
  sparse = FALSE,
  prealloc = 10000,
  nthreads = NULL
)

make_mat_gr(
  bedfiles,
  regions,
  column,
  mat_name = "value",
  prealloc = 10000,
  nthreads = NULL
)
}
\arguments{
\item{bedfiles}{A vector of BED file paths}

\item{regions}{A vector, data frame or GenomicRanges of genomic regions. See details.}

\item{column}{The index of the data column needed for the matrix}

\item{mat_name}{What to name the matrix in the returned object}

\item{sparse}{Whether to return a sparse matrix}

\item{prealloc}{The number of rows to initialize the matrices with. If the
number of loci are approximately known, this can reduce runtime as fewer
resizes need to be made.}

\item{nthreads}{Set the number of threads to use. Overrides the
\code{"iscream.threads"} option. See \code{?set_threads} for more information.}
}
\value{
\itemize{
\item \code{make_mat()}: A named list of
\itemize{
\item the matrix with the value of interest
\item a character vector of chromosomes and numeric vector of base positions
\item a character vector of the input sample BED file names
}
\item \code{make_mat_gr()}: if \code{GenomicRanges} is available, a \code{GRanges}
\item \code{make_mat_se()}: if \code{SummarizedExperiment} is available, a \code{RangedSummarizedExperiment}
}
}
\description{
Queries the provided regions and produces a matrix along with genomic
positions as a named list (\code{make_mat()}), a \code{RangedSummarizedExperiment}
(\code{make_mat_se()}), \code{GRanges} (\code{make_mat_gr()}). Parallelized across files
using threads from the \code{"iscream.threads"} option.
}
\details{
The input regions may be string vector in the form "chr:start-end"
or a GRanges object. If a data frame is provided, they must have "chr",
"start", and "end" columns.
}
\examples{
bedfiles <- system.file("extdata", package = "iscream") |>
  list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
# examine the bedfiles
colnames <- c("chr", "start", "end", "beta", "coverage")
lapply(bedfiles, function(i) knitr::kable(read.table(i, col.names = colnames)))

# make a vector of regions
regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
# make matrix of beta values
make_mat(bedfiles, regions, column = 4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarize_meth_regions.R
\name{summarize_meth_regions}
\alias{summarize_meth_regions}
\title{Summarize methylation information over genomic regions}
\usage{
summarize_meth_regions(
  bedfiles,
  regions,
  fun = "all",
  aligner = "biscuit",
  feature_col = NULL,
  mval = TRUE,
  set_region_rownames = FALSE,
  nthreads = NULL
)
}
\arguments{
\item{bedfiles}{A vector of BED file paths}

\item{regions}{A vector, data frame or GenomicRanges of genomic regions. See
details.}

\item{fun}{Function(s) to apply over the region. See details.}

\item{aligner}{The aligner used to produce the BED files - one of "biscuit",
"bismark", "bsbolt".}

\item{feature_col}{Column name of the input \code{regions} data frame containing
a name for each genomic region. Set only if the using a data frame as the
input regions format. See details.}

\item{mval}{Whether to calculate the M value (coverage \eqn{\times \beta})
or use the beta value when applying the function.}

\item{set_region_rownames}{Use the region strings as the returned data
frame's rownames. Can be useful if you have a named regions and want both
the regions strings rownames and the feature names. See details.}

\item{nthreads}{Set number of threads to use overriding the
\code{"iscream.threads"} option. See \code{?set_threads} for more information.}
}
\value{
A data.frame
}
\description{
Run summarizing functions on the CpG/CpH loci in BED files across genomic
regions. Parallelized across files using threads from the
\code{"iscream.threads"} option.
}
\section{Supported functions}{
\itemize{
\item Sum: \code{"sum"}
\item Mean: \code{"mean"}
\item Median: \code{"median"}
\item Standard deviation: \code{"stddev"}
\item Variance: \code{"variance"}
\item Minimum: \code{"min"}
\item Maximum: \code{"max"}
\item Range: \code{"range"}
\item No. of records in the region: \code{"count"}
}

The summarizing computations are backed by the Armadillo library. See
\url{https://arma.sourceforge.net/docs.html#stats_fns} for futher details on the
supported functions
}

\section{Using feature identifiers}{
\code{regions} may be string vector in the form "chr:start-end", a GRanges
object or a data frame with "chr", "start", and "end" columns. The \code{feature}
column of the output will contain a "chr:start-end" identifier for each
summarized region. To use other identifiers, like a gene name for a region
instead of the coordinates, set the names of the vector or GRanges to those
identifiers. These names will be used instead of the genomic region string
to describe each feature in the output dataframe. If \code{regions} is a data
frame make an additional column with the identifiers and pass that column
name to \code{feature_col}. See examples.
}

\examples{
# also see examples from ?summarize_regions

bedfiles <- system.file("extdata", package = "iscream") |>
  list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)

# make a vector of regions
regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
summarize_meth_regions(bedfiles, regions)
names(regions) <- c("A", "B", "C")
summarize_meth_regions(bedfiles, regions, fun = c("mean", "stddev"), mval = FALSE)
summarize_meth_regions(bedfiles, regions, fun = "sum")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{Cpp_summarize_regions}
\alias{Cpp_summarize_regions}
\title{Apply a function over BED file records within genomic features}
\usage{
Cpp_summarize_regions(
  bedfiles,
  regions,
  fun_vec,
  col_indices,
  col_names,
  aligner,
  mval = FALSE,
  region_rownames = FALSE,
  nthreads = 1L
)
}
\arguments{
\item{bedfiles}{A vector of BED file paths}

\item{regions}{A vector of genomic regions}

\item{fun_vec}{Vector of the armadillo-supported stats functions to apply over the
CpGs in the ' regions: \code{"sum"}, \code{"mean"}, \code{"median"}, \code{"stddev"},
\code{"variance"} "\code{count}", \code{"min"},\code{"max"}, and \code{"range"}.}

\item{col_indices}{A vector of genomic regions}

\item{col_names}{A vector of genomic regions}

\item{mval}{Calculates M values when TRUE, use beta values when FALSE}

\item{region_rownames}{Whether to set rownames to the regions strings. Not
necessary if your regions vector is unnamed. If its names, then the "feature"
column is set to the names and the rownames are set to the regions string}

\item{nthreads}{Number of cores to use. See details.}
}
\value{
A summary data.frame
}
\description{
This function should be called from \code{summarize_regions()} since there are few
sanity checks on the C++ side.
}
\details{
The optimal number of threads depends on the number of bedfiles, but is set
to half the available OpenMP cores. See \code{?get_threads} for more details. It
can be manaully set with \code{set_threads()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/threads.R
\name{get_threads}
\alias{get_threads}
\title{Get the number of available threads}
\usage{
get_threads()
}
\value{
A named vector:
\itemize{
\item \code{use_threads} = the number of threads iscream will use
\item \code{opt_set} = whether the option was set by the user
\item \code{avail_threads} = The number of available threads as reported by
\code{parallelly::availableCores}
}
}
\description{
Gets the number of threads iscream is currently set to use, whether the
\code{"iscream.threads"} option is set and how many threads are available for
use. To set the number of threads use \code{set_threads()} or set the
\code{iscream.threads} option in your \verb{~/.Rprofile}. See \code{?set_threads} for more
information.
}
\examples{
get_threads()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verify_files.R
\name{check_files_exist}
\alias{check_files_exist}
\title{Check that files exist}
\usage{
check_files_exist(files_vec, error_file_prefix = "Bedfile")
}
\arguments{
\item{files_vec}{A vector of file paths}

\item{error_file_prefix}{Error message prefix for 'Bedfile' vs 'Tabix file'}
}
\value{
TRUE if all input BED files have an associated tabix index file.
FALSE if not
}
\description{
Check that files exist
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verify_files.R
\name{verify_filetype}
\alias{verify_filetype}
\title{Verify that the input BED files are of the type specified by the input aligner}
\usage{
verify_filetype(bedfiles, aligner, stop_on_error = FALSE)
}
\arguments{
\item{bedfiles}{A vector of BED file paths}

\item{aligner}{The aligner chosen}

\item{stop_on_error}{Whether to warn or stop on aligner-filename mismatch}
}
\value{
TRUE if all input BED files have an associated tabix index file.
FALSE if not
}
\description{
Verify that the input BED files are of the type specified by the input aligner
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_df_string}
\alias{get_df_string}
\title{DataFrame to region strings}
\usage{
get_df_string(regions_df, feature_col = NULL)
}
\arguments{
\item{regions_df}{A data frame with "chr", "start" and "end" columns}

\item{feature_col}{The data frame column to use as the names of the output string vector}
}
\value{
A character vector
}
\description{
Convert DataFrame to a vector of strings. Set feature names in a "name" column
}
\examples{
(df <- data.frame(chr = c("chr1", "chr2"), start = c(1, 5), end = c(4, 10)))
get_df_string(df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabix.R, R/tabix_gr.R, R/tabix_raw.R
\name{tabix}
\alias{tabix}
\alias{tabix_gr}
\alias{tabix_raw}
\title{Query records from tabixed BED files}
\usage{
tabix(bedfiles, regions, aligner = NULL, col.names = NULL, nthreads = NULL)

tabix_gr(
  bedfiles,
  regions,
  aligner = NULL,
  col.names = NULL,
  zero_based = TRUE,
  nthreads = NULL
)

tabix_raw(bedfiles, regions, nthreads = NULL)
}
\arguments{
\item{bedfiles}{The BED files to be queried}

\item{regions}{A vector, data frame or GenomicRanges of genomic regions. See
details.}

\item{aligner}{The aligner used to produce the BED files - one of "biscuit",
"bismark", "bsbolt". Will set the result data.table's column names based on
this argument.}

\item{col.names}{A vector of column names for the data columns of the
result.table, not including "chr", "start", and "end". Set if your BED file
is not from the supported aligners or is a general BED file.}

\item{nthreads}{Set number of threads to use overriding the
\code{"iscream.threads"} option. See \code{?set_threads} for more information.}

\item{zero_based}{Whether the input BED file has a zero-based start column -
used when coverting the result data frame to GenomicRanges.}
}
\value{
\itemize{
\item \code{tabix()}: A data frame
\item \code{tabix_gr()}: A \code{GRanges} object for single files and \code{GRangesList} for
multiple files. When making \code{GRanges}, the 0-based records from BED-files
will be converted to 1-based with
\code{GenomicRanges::makeGRangesFromDataFrame()}. Bismark's coverage files will
not be converted as they are already 1-based and the \code{ranges} slot will be
only one position.
\item \code{tabix_raw()}: A named list of raw strings from the regions in the style
of \code{Rsamtools::scanTabix}
}
}
\description{
Query records from tabixed BED files
}
\details{
\subsection{Query method}{

'\emph{iscream} has two methods to query records from BED files:
\itemize{
\item the \emph{tabix} shell executable: fast since its output can be redirected to a
file (which \code{data.table::fread()} can then read) instead of having to
allocate memory and store it during the query
\item \emph{iscream's} tabix implementation, based on the \emph{tabix} executable using
\emph{htslib}, but slower on large queries since it stores the records as they
are found instead of writing to a file. However it's able to store each
region's records independently instead of in a single file and is used in
\code{make_mat()}, \code{make_mat_bsseq()}, and \code{summarize_regions()}.
}

When \emph{iscream} is attached, it checks that the \emph{tabix} executable is
available with \code{Sys.which()} and, if available, sets \code{options("tabix.method" = "shell")}. \code{tabix()} then uses the \emph{tabix} executable to make queries,
except for \code{tabix_raw()}. If \emph{tabix} is not found, \emph{iscream} uses its tabix
implementation. To use only \emph{iscream's} tabix implementation, set
\code{options("tabix.method" = "htslib")}.
}

\subsection{Input region formats}{

The input regions format may be string vector in the form "chr:start-end",
a dataframe with "chr", "start" and "end" columns or a \code{GRanges} object.
Input regions must be 1-based. When using \code{"htslib"} as the query method, if
the input \code{GRanges} object of regions contains any single locus regions
where the start and end positions are the same, iscream will notify that
such regions were found and fixed as \code{chr:start} format strings are invalid
for the htslib API (see \code{?get_granges_string}).
}
}
\examples{
bedfiles <- system.file("extdata", package = "iscream") |>
  list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
tabix(bedfiles, regions, col.names = c("beta", "coverage"))
if (require("GenomicRanges", quietly = TRUE)) {
  tabix_gr(bedfiles, regions, col.names = c("beta", "coverage"))
}
tabix_raw(bedfiles, regions)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R
\name{validate_log_level}
\alias{validate_log_level}
\title{Validate provided logging level}
\usage{
validate_log_level(level = get_log_level(), n_threads)
}
\arguments{
\item{level}{The logging level to validate}

\item{n_threads}{The number of threads that the next iscream function call
will use}
}
\value{
None; sets the log level to the provide level
}
\description{
Only "info" and "debug" are currently supported, with "debug" only supported
when using 1 thread
}
\examples{
set_log_level("info")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/threads.R
\name{check_thread_count}
\alias{check_thread_count}
\title{Check that the required threads are available}
\usage{
check_thread_count(
  n_threads,
  avail_threads = availableCores(),
  opt_set = FALSE
)
}
\arguments{
\item{n_threads}{The number of threads to check availability for}

\item{avail_threads}{The number of threads that are available on the system.
Defaults to \code{parallelly::availableCores()}}

\item{opt_set}{Whether the \code{iscream.threads} options is set}
}
\value{
\code{n_threads} if the requested number of threads are available and
stops if not
}
\description{
Check that the required threads are available
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarize_regions.R
\name{summarize_regions}
\alias{summarize_regions}
\title{Summarize information over genomic regions from any BED file}
\usage{
summarize_regions(
  bedfiles,
  regions,
  columns,
  col_names = NULL,
  fun = "all",
  feature_col = NULL,
  set_region_rownames = FALSE,
  nthreads = NULL
)
}
\arguments{
\item{bedfiles}{A vector of BED file paths}

\item{regions}{A vector, data frame or GenomicRanges of genomic regions. See
details.}

\item{columns}{A vector of indices of the numeric columns to be summarized}

\item{col_names}{A vector of names to use for \code{columns} in the output}

\item{fun}{Function(s) to apply over the region. See details.}

\item{feature_col}{Column name of the input \code{regions} data frame containing
a name for each genomic region. Set only if the using a data frame as the
input regions format. See details.}

\item{set_region_rownames}{Use the region strings as the returned data
frame's rownames. Can be useful if you have a named regions and want both
the regions strings rownames and the feature names. See details.}

\item{nthreads}{Set number of threads to use overriding the
\code{"iscream.threads"} option. See \code{?set_threads} for more information.}
}
\value{
A data.frame
}
\description{
Run summarizing functions on BED file records across genomic regions.
Parallelized across files using threads from the \code{"iscream.threads"} option.
}
\section{Supported functions}{
\itemize{
\item Sum: \code{"sum"}
\item Mean: \code{"mean"}
\item Median: \code{"median"}
\item Standard deviation: \code{"stddev"}
\item Variance: \code{"variance"}
\item Minimum: \code{"min"}
\item Maximum: \code{"max"}
\item Range: \code{"range"}
\item No. of records in the region: \code{"count"}
}

The summarizing computations are backed by the Armadillo library. See
\url{https://arma.sourceforge.net/docs.html#stats_fns} for futher details on the
supported functions
}

\section{Using feature identifiers}{
\code{regions} may be string vector in the form "chr:start-end", a GRanges
object or a data frame with "chr", "start", and "end" columns. The \code{feature}
column of the output will contain a "chr:start-end" identifier for each
summarized region. To use other identifiers, like a gene name for a region
instead of the coordinates, set the names of the vector or GRanges to those
identifiers. These names will be used instead of the genomic region string
to describe each feature in the output dataframe. If \code{regions} is a data
frame make an additional column with the identifiers and pass that column
name to \code{feature_col}. See examples.
}

\examples{
bedfiles <- system.file("extdata", package = "iscream") |>
  list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
# examine the bedfiles
colnames <- c("chr", "start", "end", "beta", "coverage")
lapply(bedfiles, function(i) knitr::kable(read.table(i, col.names = colnames)))

# make a vector of regions
regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
summarize_regions(bedfiles, regions, columns = c(4, 5), col_names = c("beta", "cov"))

# select functions
summarize_regions(
  bedfiles,
  regions,
  fun = c("mean", "stddev"),
  columns = c(4, 5),
  col_names = c("beta", "cov")
)

# add names to the regions
names(regions) <- c("A", "B", "C")
summarize_regions(
  bedfiles,
  regions,
  fun = "sum",
  columns = 5,
  col_names = "coverage"
)

# using `feature_col`
library(data.table)

# convert string vector to a data.table
regions_df <- data.table::as.data.table(regions) |>
_[, tstrsplit(regions, ":|-", fixed = FALSE, names = c("chr", "start", "end"))] |>
_[, start := as.integer(start)] |>
_[, feature := LETTERS[.I]][]
regions_df

summarize_regions(
  bedfiles,
  regions_df,
  fun = "sum",
  columns = 5,
  col_names = "coverage",
  feature_col = "feature"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{htslib_version}
\alias{htslib_version}
\title{Get htslib version and available features}
\usage{
htslib_version()
}
\value{
None
}
\description{
Returns the version of htslib being used by iscream and whether features
such as libdeflate support are available. This information may not always
correspond to the htslib version used during iscream's installation if a
different htslib version is available for linking at runtime.
}
\examples{
htslib_version()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{verify_aligner_or_stop}
\alias{verify_aligner_or_stop}
\title{Validate provided aligner}
\usage{
verify_aligner_or_stop(aligner)
}
\arguments{
\item{aligner}{The input alinger}
}
\value{
true; quits if the input is not among supported_aligners
}
\description{
Only "biscuit", "bismark", and "bsbolt" are currently supported
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_granges_string}
\alias{get_granges_string}
\title{GRanges to region strings}
\usage{
get_granges_string(gr)
}
\arguments{
\item{gr}{A GRanges object}
}
\value{
A character vector
}
\description{
Coerces GenomicRanges to \code{chr:start-end} strings with \code{as.character}. If any
regions have the same start and end, \code{as.character} returns \code{chr:start}
strings which are invalid for the htslib API. These are corrected to
\code{chr:start-start}.
}
\examples{
if (requireNamespace("GenomicRanges", quietly = TRUE)) {
  get_granges_string(GenomicRanges::GRanges(c("chr1:1-10", "chr2:15-20")))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/verify_files.R
\name{verify_files_or_stop}
\alias{verify_files_or_stop}
\title{Verify that BED files are tabixed}
\usage{
verify_files_or_stop(bedfiles, verify_tabix = TRUE)
}
\arguments{
\item{bedfiles}{A vector of BED file paths}

\item{verify_tabix}{Whether to verify the presence of tabix files}
}
\value{
TRUE if all input BED files have an associated tabix index file.
FALSE if not
}
\description{
Verify that BED files are tabixed
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/threads.R
\name{set_threads}
\alias{set_threads}
\title{Set the number of available threads}
\usage{
set_threads(n_threads)
}
\arguments{
\item{n_threads}{The number of threads to use}
}
\value{
None. Sets the \code{iscream.threads} option to the requested number of
threads if available
}
\description{
Sets the \code{"iscream.threads"} option to \code{n_threads}. To see how many threads
you have available see \code{?get_threads()}.
}
\details{
iscream uses OpenMP to parallelize certain functions. You can use
as many threads as are available to you on your system to varying degrees of
performance improvements. The \code{get_threads()} function uses
\code{parallelly::availableCores()} to report the number of available threads.
Although OpenMP can detect the number of available cores, on high
performance computers (HPCs) with resource allocating job schedulers like
SLURM, OpenMP may detect all available threads across the HPC and not limit
itself to the cores that were allocated to you by the scheduler. If your
system administrator has not set up any limits, this may result in your job
taking resources from other jobs. If there are limits, trying to use more
threads that those available will reduce iscream's performance. Job
schedulers will typically have an environment variable (e.g.
\code{SLURM_CPUS_ON_NODE} with SLURM) that gives you the actual number of
available cores. Further, on hyperthreaded systems, this count may be double
that of the available processors. Using hyperthreading does not guarantee
any performance improvement - it may be better to set the number of threads
to half the reported number. \code{parallelly::availableCores()} takes HPC
scheduler/CRAN/Bioconductor limits into account when reporting the number of
available threads but it may not reliably report hyperthreading ('system' or
'nproc'). To set the number of threads without having to call
\code{set_threads()} in every session, put

\if{html}{\out{<div class="sourceCode">}}\preformatted{options(iscream.threads = [n_threads])
}\if{html}{\out{</div>}}

in your \code{.Rprofile} See \code{help('Rprofile')} for information on startup options.

Functions currently using multithreading:
\itemize{
\item \code{tabix()}, \code{tabix_gr()}, \code{tabix_raw()}
\item \code{query_chroms()}
\item \code{make_mat()}, \code{make_mat_se()}, \code{make_mat_gr()}, \code{make_mat_bsseq()}
\item \code{summarize_regions()}, \code{summarize_meth_regions()}
}
}
\examples{
(ncores <- parallelly::availableCores())
\donttest{
set_threads(ncores)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{Cpp_set_log_level}
\alias{Cpp_set_log_level}
\title{spdlog Logging Lever Setter}
\usage{
Cpp_set_log_level(name)
}
\arguments{
\item{name}{A string with the logging level. Value understood are,
in decreasing verbosity \sQuote{trace}, \sQuote{debug}, \sQuote{info},
\sQuote{warning}, \sQuote{error}, \sQuote{critical}, and \sQuote{off}.
Unrecognised names are equivalent to \sQuote{off}.}
}
\value{
Nothing is returned.
}
\description{
A helper function to turn a logging level given as string
into the current logging level
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{query_chroms}
\alias{query_chroms}
\title{Query the chromosomes or seqnames from a vector of BED files}
\usage{
query_chroms(bedfiles, nthreads = NULL)
}
\arguments{
\item{bedfiles}{The vector of BED file paths}

\item{nthreads}{Set number of threads to use overriding the
\code{"iscream.threads"} option. See \code{?set_threads} for more information.}
}
\value{
A vector of seqnames
}
\description{
Query the chromosomes or seqnames from a vector of BED files
}
\examples{
bedfiles <- system.file("extdata", package = "iscream") |>
  list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
query_chroms(bedfiles)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log.R, R/RcppExports.R
\name{set_log_level}
\alias{set_log_level}
\alias{get_log_level}
\title{Set and get logging level}
\usage{
set_log_level(level = "info")

get_log_level()
}
\arguments{
\item{level}{The logging verbosity level to use
\itemize{
\item \code{"info"}: the default that gives provides basic information about the
number of files and regions used in a function
\item \code{"debug"}: more verbose about row allocations, how many CpGs were found in
a region, filename parsing etc. This mode cannot be used on more than one
thread as R cannot output messages from multiple threads without crashing.
\item \code{"off"}: no logging
}}
}
\value{
\itemize{
\item \code{set_log_level()}: None; sets the log level to the provided level
\item \code{get_log_level()}: The current logging level as a string
}
}
\description{
Set and get logging level
}
\examples{
set_log_level("info")
get_log_level()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/make_mat_bsseq.R
\name{make_mat_bsseq}
\alias{make_mat_bsseq}
\title{Make M/beta and coverage matrices from WGBS BED files}
\usage{
make_mat_bsseq(
  bedfiles,
  regions,
  aligner = "biscuit",
  mval = TRUE,
  merged = TRUE,
  sparse = FALSE,
  prealloc = 10000,
  nthreads = NULL
)
}
\arguments{
\item{bedfiles}{A vector of BED file paths}

\item{regions}{A vector, data frame or GenomicRanges of genomic regions. See details.}

\item{aligner}{The aligner used to produce the BED files - one of "biscuit",
"bismark", "bsbolt".}

\item{mval}{Whether to return M-values or beta-values with the coverage
matrix. Defaults to M-value. Set \code{mval=FALSE} to get beta value matrix.}

\item{merged}{Whether the input strands have been merged/collapsed}

\item{sparse}{Whether to return a sparse matrix}

\item{prealloc}{The number of rows to initialize the matrices with. If the
number of loci are approximately known, this can reduce runtime as fewer
resizes need to be made.}

\item{nthreads}{Set the number of threads to use. Overrides the
\code{"iscream.threads"} option. See \code{?set_threads} for more information.}
}
\value{
A named list of
\itemize{
\item coverage and either a beta- or M-value matrix
\item a character vector of chromosomes and numeric vector of corresponding CpG
base positions
\item a character vector of the input sample names
}
}
\description{
Queries the CpG/CpH loci from provided regions and produces M/beta and
coverage matrices with their genomic positions. Parallelized across files
using threads from the \code{"iscream.threads"} option. The output of
\code{make_mat_bsseq} may be used to create a BSseq object: \code{do.call(BSseq, make_mat_bsseq(...))}.
}
\details{
The input regions may be string vector in the form "chr:start-end"
or a GRanges object. If a data frame is provided, they must have "chr",
"start", and "end" columns.
}
\section{Bitpacking limits}{

\code{make_mat_bsseq()} makes two matrices: M-value (or beta-value) and coverage.
For speed and memory efficiency these two values are bitpacked during matrix
creation so that only one matrix needs to be populated and resized. This
matrix is unpacked into the two required matrices only after the matrix
dimensions are known after querying all input files. The two values are
packed using the INT16 type, which has an upper limit of 32,767, into one
INT32. If the coverage values exceed 32,767, the upper limit of a 16-bit
signed integer, it will be capped at the limit. Beta values will also be
capped similarly, but any such beta values would indicate a bug in the
aligner that produced the data.
}

\examples{
bedfiles <- system.file("extdata", package = "iscream") |>
  list.files(pattern = "[a|b|c|d].bed.gz$", full.names = TRUE)
# examine the BED files
colnames <- c("chr", "start", "end", "beta", "coverage")
lapply(bedfiles, function(i) knitr::kable(read.table(i, col.names = colnames)))

# make a vector of regions
regions <- c("chr1:1-6", "chr1:7-10", "chr1:11-14")
mat <- make_mat_bsseq(bedfiles, regions)
# for BSseq object run
if (requireNamespace("bsseq", quietly = TRUE)) {
  do.call(bsseq::BSseq, mat)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{Cpp_query_all}
\alias{Cpp_query_all}
\title{Query all methylation info into M and coverage matrices}
\usage{
Cpp_query_all(
  bedfiles,
  regions,
  aligner,
  valInd,
  merged,
  sparse,
  prealloc,
  nthreads
)
}
\arguments{
\item{bedfiles}{A vector of BED files}

\item{regions}{A vector of regions}

\item{aligner}{The aligner used to make the WGBS BED files, only for
\code{make_mat_bsseq}}

\item{valInd}{The index of the data column needed for the matrix, for \code{make_mat}}

\item{merged}{Whether the input strands have been merged/collapsed}

\item{prealloc}{The number of rows to initialize the matrices with}

\item{nthreads}{Set number of threads to use overriding the
\code{"iscream.threads"} option. See \code{?set_threads} for more information.}
}
\value{
A list of one or two matrices, chromosome, position, and filename vectors
}
\description{
Query all methylation info into M and coverage matrices
}
\keyword{internal}
